import math
import copy
from .. import constants
from .. import tree
from .. import shape
from .. import solid
from .. import boolean
from .. import transform
from ..attributes import Attributes
from ..util import differentialdrawing
from ..util import matrix
from ..util import hash
from ..util import cache



class Geometry:
    """
    Basic geometry of a gear part.
    """
    def __init__ (self,
                    n = 20,                         # number of tooth               (de: Zahnanzahl)
                    m = None,                       # module                        (de: Modul)
                    d0 = None,                      # base circle diameter          (de: Teilkreisdurchmesser)
                    da = None,                      # addendum circle diameter      (de: Kopfkreisdurchmesser)
                    p0 = None,                      # circular pitch                (de: Umfangsteilung)
                    x = None,                       # profile shift factor          (de: Profilverschiebungsfaktor)
                    a = 20,                         # standard pressure angle       (de: Normaleingriffswinkel)
                    b = 0,                          # helix angle                   (de: Schrägungswinkel)
                    innerGear = False               # create shape for inner gears  (de: Umriss für innenverzahnung erzeugen)
        ):

        #: Number of tooth.
        self.n = n

        # calculate m, d0, da and p0
        if m != None:
            assert p0 == None, "parameter p0 is not accepted when parameter m is used"
            assert d0 == None, "parameter d0 is not accepted when parameter m is used"
            assert da == None, "parameter da is not accepted when parameter m is used"
            _m = m
            _p0 = math.pi * m
            _d0 = _m * self.n
            _da = _m * (self.n + 2)
        else:
            if (p0 != None):
                assert d0 == None, "parameter d0 is not accepted when parameter p0 is used"
                assert da == None, "parameter da is not accepted when parameter p0 is used"
                _m = p0 / math.pi
                _p0 = p0
                _d0 = _m * self.n
                _da = _m * (self.n + 2)
            else:
                if d0 != None:
                    assert da == None, "parameter da is not accepted when parameter d0 is used"
                    _m = d0 / self.n
                    _p0 = _m * math.pi
                    _d0 = d0
                    _da = _m * (self.n + 2)
                elif da != None:
                    _m = da / (self.n + 2)
                    _p0 = _m * math.pi
                    _d0 = _m * self.n
                    _da = da
                else:
                    pStr = "n = " + str (n) + ", "
                    pStr += "m = " + str (m) + ", "
                    pStr += "d0 = " + str (d0) + ", "
                    pStr += "da = " + str (da) + ", "
                    pStr += "p0 = " + str (p0) + ", "
                    pStr += "x = " + str (x) + ", "
                    pStr += "a = " + str (a) + ", "
                    pStr += "b = " + str (b)
                    assert False, "can not calculate parameter m: " + pStr

        #: Module of gear.
        self.m = _m

        #: Circular pitch.
        self.p0 = _p0

        #: Base circle diameter.
        self.d0 = _d0

        #: Addendum circle diameter.
        self.da = _da

        # profile shift
        if x == None:
            if self.n < 16:
                _x = 1 - (self.n / 17)
            else:
                _x = 0
        else:
            _x = x

        #: Profile shifting factor.
        self.x = _x

        # TODO: rewirte values alpha and beta to be in degree.
        degToRad = math.pi / 180.0

        _alpha = 20 * degToRad
        if a != None:
            _alpha = a * degToRad

        #: Standard pressure angle.
        self.a = _alpha

        _beta = 0
        if b != None:
            _beta = b * degToRad

        #: Helix angle.
        self.b = _beta

        # Create geometry for inner gear
        self.innerGear = innerGear


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            'n',
            'm',
            'x',
            'a',
            'b'
        )



# TODO: make methods and variables private
# TODO: add base class shape.Shape (and add function getDimensions)
class Profile (shape.Complex, Geometry):
    """
    Profile of a gear wheel.
    """
    def __init__(self, 
                    n = 20,                         # number of tooth               (de: Zahnanzahl)
                    m = None,                       # module                        (de: Modul)
                    d0 = None,                      # base circle diameter          (de: Teilkreisdurchmesser)
                    da = None,                      # addendum circle diameter      (de: Kopfkreisdurchmesser)
                    p0 = None,                      # circular pitch                (de: Umfangsteilung)
                    x = None,                       # profile shift factor          (de: Profilverschiebungsfaktor)
                    a = 20,                         # standard pressure angle       (de: Normaleingriffswinkel)
                    b = 0,                          # helix angle                   (de: Schrägungswinkel)
                    mhf = 0.2,                      # factor of minimum tooth width (de: Faktor für minimale Zahnkopfbreite)
                    rot = 0,                        # rotate gear wheel             (de: Rotation des Zahnrads)
                    innerGear = False,              # create shape for inner gears  (de: Umriss für innenverzahnung erzeugen)
                    name: str = None,               # name of shape
                    attributes:Attributes = None,   # attributes of shape
        ):
        """
        Creates a basic gear wheel geometrie.

        Possible combination of parameters:
        n & m,
        n & d0,
        n & da,
        n & p0

        All other parameters will be calculated
        """
        shape.Complex.__init__ (self, name = name, attributes = attributes)
        Geometry.__init__ (
            self,
            n = n,
            m = m,
            d0 = d0,
            da = da,
            p0 = p0,
            x = x,
            a = a,
            b = b
        )

        #: Boolean flag: is inner gear.
        self.innerGear = innerGear

        # Clearance value.
        self.c = 0.167 * self.m

        #: Rotation of gear wheel in degree.
        self.rot = rot

        # Absolute helix angle.
        betaAbs = self.b if self.b > 0 else -self.b

        #: Base circle diameter
        self.db = self.d0 * math.cos (math.atan (math.tan (self.a) / math.cos (betaAbs)))

        # Precalculated profile shift offset
        self.v = self.x * self.m

        # Correct addendum circle with profile shift
        self.da += 2 * self.v

        # Tip shortening
        maxDa = (self._centerDistanceToNonCorrected () - ((self.d0 / 2) - self.m)) * 2
        if self.da > maxDa:
            self.da = maxDa

        headTopWidth = self._toothWidthAtDiameter (self.da) / 2
        self.mhf = mhf
        if headTopWidth < (mhf * self.m):
            self.da = self._findDiameterForWidth (self.d0, self.da, mhf * self.m * 2)

        # Pre calculations
        # TODO: make names private
        self.rb = self.db / 2
        self.r0 = self.d0 / 2
        self.ra = self.da / 2
        self.p0 = math.pi * self.m
        self.rho_ra = math.acos (self.rb / self.ra)
        self.rho_r0 = math.acos (self.rb / self.r0)
        self.headAngleExtend = self._toothAngleAtDiameter (self.da)
        self.toothAngleAtD0 = self._toothAngleAtDiameter (self.d0)
        self.toothAngleAtBase = self._toothAngleAtDiameter (self.db)


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            shape.Complex,
            Geometry,
            'mhf',
            'innerGear',
            'rot',
            'name',
            'attributes'
        )


    def _involute (self, a):
        """
        Involute function.
        """
        return math.tan (a) - a


    def _solveInvolute (self, invA):
        """
        Solve involute function.
        """
        if (invA < -2) or (invA > 2):
            aLast = math.atan (invA)
        else:
            aLast = math.pow (3 * invA, 1.0 / 3.0) - (2.0 / 5.0) * invA
        maxIterations = 50
        while (maxIterations > 0):
            at = math.tan (aLast)
            aNext = aLast + ((invA - at + aLast) / (at * at))
            absError = aNext - aLast
            if absError < 0:
                absError = -absError
            if absError < 0.00000000001:
                return aNext
            aLast = aNext
            maxIterations -= 1
        assert False, "solver not converging"


    def centerDistance (self, other):
        """
        Calculates the center distance between two gear wheels.
        """
        assert (self.m == other.m), "gear wheels must have the same module"
        assert (self.a == other.a), "gear wheels must have the same module"
        if self.innerGear and other.innerGear:
            assert False, "can not pair two inner gear wheels"
        if other.innerGear:
            return self._centerDistanceWithInner (other)
        if self.innerGear:
            return other._centerDistanceWithInner (self)
        assert (self.b == -other.b), "gear wheels must have opposit helix angles"
        return self._centerDistanceNonInner (other)


    def _centerDistanceNonInner (self, other):
        invAb = 2 * ((self.x + other.x) / (self.n + other.n)) * math.tan (self.a) + self._involute (self.a)
        ab = self._solveInvolute (invAb)
        return self.m * (self.n + other.n) * (math.cos (self.a) / (2 * math.cos (ab)))


    def _centerDistanceWithInner (self, inner):
        """
        Calculate the center distance between a gear and an inner gear.
        """
        if inner.x < self.x:
            assert False, "inner gear must have at least the same profile shift factor than gear"
        innerDist = inner._centerDistanceNonInner (inner)
        gearDist = self._centerDistanceNonInner (inner)
        return innerDist - gearDist


    def _centerDistanceToNonCorrected (self):
        """
        Calculates the center distance between this gear wheel and non corrected gear wheel with same parameters.
        """
        invAb = 2 * (self.x / (2 * self.n)) * math.tan (self.a) + self._involute (self.a)
        ab = self._solveInvolute (invAb)
        return self.m * (2 * self.n) * (math.cos (self.a) / (2 * math.cos (ab)))


    def _toothWidthAtDiameter (self, diameter):
        """
        Calculate width of tooth for a diameter.
        """
        ac = (self.d0 / diameter) * math.cos (self.a)
        if ac > 1:
            ac = 1
        a = math.acos (ac)
        invA = self._involute (a)
        invA0 = self._involute (self.a)
        x = self.x
        s0 = self.m * ((math.pi / 2) + (2 * x * math.tan (self.a)))
        return diameter * ((s0 / self.d0) + invA0 - invA)


    def _findDiameterForWidth (self, min, max, width):
        """
        Find the gear wheels diameter for a tooth width.
        """
        center = (min + max) / 2
        if (max - min) < 0.0000000000001:
            return center
        d = self._toothWidthAtDiameter (center)
        if d < width:
            return self._findDiameterForWidth (min, center, width)
        else:
            return self._findDiameterForWidth (center, max, width)


    def _toothAngleAtDiameter (self, diameter):
        """
        Calculate angle of tooth at a diameter.
        """
        if self.innerGear:
            if diameter < self.d0:
                mirroredDiameter = 2 * self.d0 - diameter
                a = (self._toothWidthAtDiameter (mirroredDiameter) / 2) / (mirroredDiameter / 2)
                return (2 * self.toothAngleAtD0 - a)
        else:
            if diameter < self.db:
                mirroredDiameter = 2 * self.db - diameter
                a = (self._toothWidthAtDiameter (mirroredDiameter) / 2) / (mirroredDiameter / 2)
                return (2 * self.toothAngleAtBase - a)
        return (self._toothWidthAtDiameter (diameter) / 2) / (diameter / 2)


    def _toothFlankDDA (self, diameter, leftSide, rotate = 0):
        """
        Caclulate evolvent point for a diameter.
        """
        r = diameter / 2
        a = self._toothAngleAtDiameter (diameter)
        return (
            -math.sin (a + rotate) * r if leftSide else math.sin (a + rotate) * r,
            math.cos (a + rotate) * r
        )


    def _undercut (self, angle, mirror = False):
        """
        Calculate undercut point for an angle.
        """
        toothWidth = (self.p0 / 2) - ((self.m + self.c / 2) * math.tan (self.a) * 2)
        toothWidthHalf = toothWidth / 2
        abroll = self.r0 * angle
        if angle > 0:
            cpX = abroll + toothWidthHalf
        else:
            cpX = abroll - toothWidthHalf
        cpY = self.r0 + self.v - (self.m + self.c / 2)
        cp = (-cpX, cpY)
        rot = matrix.AffineRotate2D (-angle * 180 / math.pi)
        rp = matrix.transformPoint2D (rot, cp)
        if mirror:
            return (-rp[0], rp[1])
        else:
            return rp


    def _undercutToEvolventDist (self, angle):
        """
        Calculate distance between undercut and evolvent for an angle.
        """
        up = self._undercut (angle)
        distFromCenter = math.sqrt (up[0] * up[0] + up[1] * up[1])
        upAngle = math.atan2 (up[0], up[1])
        toothAngle = self._toothAngleAtDiameter (distFromCenter * 2) - math.pi / self.n
        angleDiff = upAngle - toothAngle
        return angleDiff * distFromCenter


    def _findUndercutEvolventCuttingAngleNearest (self, min = 0, max = math.pi):
        """
        Find nearest cutting point between undercut and evolvent.
        """
        center = (max + min) / 2
        if (max - min) < 0.0000000000001:
            return center
        distLeft = self._undercutToEvolventDist ((min + center) / 2)
        distRight = self._undercutToEvolventDist ((max + center) / 2)
        if distLeft < distRight:
            return self._findUndercutEvolventCuttingAngleNearest (min, center)
        else:
            return self._findUndercutEvolventCuttingAngleNearest (center, max)


    def _findUndercutEvolventCuttingAngle (self, min = 0, max = math.pi):
        """
        Find cutting point between undercut and evolvent.
        """
        center = (max + min) / 2
        dist = self._undercutToEvolventDist (center)
        if (max - min) < 0.0000000000001:
            absDist = dist if dist > 0 else -dist
            if absDist > 0.000000001:
                return None
            return center
        if dist > 0:
            return self._findUndercutEvolventCuttingAngle (min, center)
        else:
            return self._findUndercutEvolventCuttingAngle (center, max)


    def _calculateUndercutAngle (self, phi):
        """
        Calculate undercut line angle for a angle on the gear wheel.
        """
        p1 = self._undercut (phi + 0.0000000000001, True)
        p2 = self._undercut (phi + 0.000000001, True)
        dx = p2[0] - p1[0]
        dy = p2[1] - p1[1]
        return math.atan2 (dy, dx)


    def _findUndercutInnerAngleLimit (self, min, max):
        """
        Find angle on gear wheel where the undercut line sould switch to a circle.
        """
        center = (min + max) / 2
        if (max - min) < 0.0000000000001:
            return center
        angle = self._calculateUndercutAngle (center)
        if angle < (math.pi / 4):
            return self._findUndercutInnerAngleLimit (center, max)
        else:
            return self._findUndercutInnerAngleLimit (min, center)


    def _generateDDA(self):
        """
        Create a differential drawing analyzer for a single tooth.
        """
        # build gear wheel or inner gear wheel
        if self.innerGear:
            toothStartDiameter = self.d0 - self.m
        else:
            # calculate undercut start position
            undercutAngle = self._findUndercutEvolventCuttingAngle ()
            if undercutAngle == None:
                undercutAngle = self._findUndercutEvolventCuttingAngleNearest ()
            assert undercutAngle != None, "undercut doesn't cross evolvent"
            undercutBeginPoint = self._undercut (undercutAngle)
            toothStartDiameter = math.sqrt (undercutBeginPoint[0] * undercutBeginPoint[0] + undercutBeginPoint[1] * undercutBeginPoint[1]) * 2
            undercutInnerAngleLimit = self._findUndercutInnerAngleLimit (0, undercutAngle)

            # create undercut
            undercutDDARight = differentialdrawing.Element(undercutAngle, undercutInnerAngleLimit, lambda a: self._undercut (a, True))
            undercutDDALeft = differentialdrawing.Element(-undercutInnerAngleLimit, -undercutAngle, lambda a: self._undercut (a, True))

            # create tooth clearance
            toothClearanceAnchestorAngle = self._calculateUndercutAngle (undercutInnerAngleLimit)

            # width of clearance area
            clearancePoint1 = undercutDDARight.pointAt (1)
            clearancePoint2 = undercutDDALeft.pointAt (0)
            tcDX = clearancePoint1[0] - clearancePoint2[0]
            tcDY = clearancePoint1[1] - clearancePoint2[1]
            toothClearanceAnchestorWidth = math.sqrt (tcDX * tcDX + tcDY * tcDY)
            toothClearanceDDA = _ToothClearanceDDA (self, toothClearanceAnchestorAngle, toothClearanceAnchestorWidth, clearancePoint1[1])

        toothEndDiameter = self.da
        toothRot = math.pi / self.n

        # create tooth with head
        leftFlankDDA = differentialdrawing.Element (toothEndDiameter, toothStartDiameter, lambda d: self._toothFlankDDA (d, True, -toothRot))
        rightFlankDDA = differentialdrawing.Element (toothStartDiameter, toothEndDiameter, lambda d: self._toothFlankDDA (d, False, toothRot))
        if self.innerGear:
            # create head clearance (inverse)
            innerCleareanceBeginPoint = leftFlankDDA.pointAt (0)
            innerCleareanceEndPoint = rightFlankDDA.pointAt (1)
            icDX = innerCleareanceBeginPoint[0] - innerCleareanceEndPoint[0]
            icDY = innerCleareanceBeginPoint[1] - innerCleareanceEndPoint[1]
            innerClearanceAnchestorWidth = math.sqrt (icDX * icDX + icDY * icDY)
            innerCleareanceAngleP1 = self._toothFlankDDA (self.da, True)
            innerCleareanceAngleP2 = self._toothFlankDDA (self.da - 0.000000001, True)
            innerClearanceAnchestorDx = innerCleareanceAngleP1[0] - innerCleareanceAngleP2[0]
            innerClearanceAnchestorDy = innerCleareanceAngleP1[1] - innerCleareanceAngleP2[1]
            innerClearanceAnchestorAngle = math.atan2 (innerClearanceAnchestorDy, innerClearanceAnchestorDx)
            headDDA = _InnerToothClearanceDDA(self, innerClearanceAnchestorAngle, innerClearanceAnchestorWidth, innerCleareanceAngleP1[1], toothRot)
        else:
            headDDA = differentialdrawing.Element (self.headAngleExtend, -self.headAngleExtend, lambda a: (math.sin (a + toothRot) * self.ra, math.cos (a + toothRot) * self.ra))

        # rasterize points of a single tooth
        if self.innerGear:
            innerRadius = self.r0 - (self.m / 2)
            innerBeginAngle = self._toothAngleAtDiameter (innerRadius * 2) - toothRot
            innerToothBase = differentialdrawing.Element (-innerBeginAngle, innerBeginAngle, lambda a: (math.sin (a) * innerRadius, math.cos (a) * innerRadius))
            dda = differentialdrawing.Path ([rightFlankDDA, headDDA, leftFlankDDA, innerToothBase])
        else:
            dda = differentialdrawing.Path ([rightFlankDDA, headDDA, leftFlankDDA, undercutDDARight, toothClearanceDDA, undercutDDALeft])
        return dda


    def _rasterize(self, resolvedAttributes, asBezier):

        """
        Create rasterized shape of gear wheel.
        """
        # get rasterizing attributes
        minAngle, minSize, fixedCount, minSlices = resolvedAttributes.rasterizingAttributes ()

        # compute element hash and try to load from cache
        ownHash = hash ("rasterized-gearwheel-profile", self, minSize, asBezier)
        cached = cache.load (ownHash, 'item')
        if cached != None:
            return cached

        # create DDA and rasterize element
        ddas = self._generateDDA ()
        if asBezier:
            toothPoints = ddas._rasterizeBezier (2, minSize, False)
        else:
            toothPoints = ddas._rasterize (2, minSize)

        # copy for each tooth
        points = []
        for tooth in range (0, self.n):
            rotAngle = (360.0 / self.n) * tooth + self.rot
            if rotAngle < 180:
                rotAngle = rotAngle + 360
            rotateMatrix = matrix.AffineRotate2D (rotAngle)
            if (len (points) > 0) and asBezier:
                # create closing element between last tooth and current one
                closingEndCurve = []
                for pindex in range (0, 4):
                    closingEndCurve.append (matrix.transformPoint2D(rotateMatrix, toothPoints[pindex]))
                closingCurve = differentialdrawing.Path._createConnectionBezier (points, closingEndCurve)
                if closingCurve != None:
                    points = points + closingCurve
            for pindex in range (0, len (toothPoints)):
                # process current tooth
                points.append (matrix.transformPoint2D(rotateMatrix, toothPoints[pindex]))
        if asBezier:
            # close curve
            closingCurve = differentialdrawing.Path._createConnectionBezier (points, points)
            if closingCurve != None:
                points = points + closingCurve

        # TODO: need to copy attributes?
        if asBezier:
            # TODO: get curves in right format
            assert False
            curves = None
            generated = shape.Bezier (curves)
        else:
            generated = shape.Polygon (points = points, name = self.name, attributes = self.attributes)

        # store to cache an return
        cache.store (generated, ownHash, 'item')
        return generated


    def _toPolygon (self, rasterizingAttributes: Attributes):
        """
        Rasterize gear profile to Polygon.
        """
        return self._rasterize (rasterizingAttributes, False)


    def _toBezier (self, rasterizingAttributes: Attributes):
        """
        Rasterize gear profile to Bezier shape.
        """
        return self._rasterize (rasterizingAttributes, True)




class _ToothClearanceDDA (differentialdrawing.Element):
    """
    DDA for creating tooth clearance.
    """
    def __init__(self, gearwheel, anchestorAngle, anchestorWidth, anchestorDistance):
        super ().__init__ (-1, 1)
        self.gearwheel = gearwheel
        self.radius = anchestorWidth / (2 * math.sin (anchestorAngle))
        self.anchestorAngle = anchestorAngle
        self.centerDist = anchestorDistance + math.cos (anchestorAngle) * self.radius


    def pointAt (self, t):
        nt = (t - 0.5) * 2
        at = -self.anchestorAngle * nt
        distFromZero = nt
        if distFromZero < 0:
            distFromZero = -distFromZero
        yt = math.cos (at) * self.radius
        if yt < 0:
            yt = -yt
        y = self.centerDist - yt
        x = -math.sin (at) * self.radius
        return (-x, y)




class _InnerToothClearanceDDA (differentialdrawing.Element):
    """
    DDA for creating tooth clearance for inner gears.
    """
    def __init__(self, gearwheel, anchestorAngle, anchestorWidth, anchestorDistance, toothRot):
        super ().__init__ (-1, 1)
        self.gearwheel = gearwheel
        self.radius = anchestorWidth / (2 * math.sin (anchestorAngle))
        self.anchestorAngle = anchestorAngle
        self.centerDist = anchestorDistance - math.cos (self.anchestorAngle) * self.radius
        self.rotMatrix = matrix.AffineRotate2D (-toothRot * 180 / math.pi)


    def pointAt (self, t):
        nt = -(t - 0.5) * 2
        at = nt * self.anchestorAngle
        atAbs = at if at > 0 else -at
        cp = (
            math.sin (at) * self.radius,
            math.cos (atAbs) * self.radius + self.centerDist
        )
        return matrix.transformPoint2D (self.rotMatrix, cp)




class Wheel (solid.LinearExtrude, Geometry):
    """
    Gear wheel as :ref:`Solid<Solid>`.
    """
    def __init__ (self,
                    n = 20,                         # number of tooth               (de: Zahnanzahl)
                    m = None,                       # module                        (de: Modul)
                    height = 1,                     # height of wheel               (de: Hoehe)
                    d0 = None,                      # base circle diameter          (de: Teilkreisdurchmesser)
                    da = None,                      # addendum circle diameter      (de: Kopfkreisdurchmesser)
                    p0 = None,                      # circular pitch                (de: Umfangsteilung)
                    x = None,                       # profile shift factor          (de: Profilverschiebungsfaktor)
                    a = 20,                         # standard pressure angle       (de: Normaleingriffswinkel)
                    b = 0,                          # helix angle                   (de: Schrägungswinkel)
                    mhf = 0.2,                      # factor of minimum tooth width (de: Faktor für minimale Zahnkopfbreite)
                    rot = 0,                        # rotate gear wheel             (de: Rotation des Zahnrads)
                    children: list = None,          # shape of gearwheel, only for deserialization, will be generated otherwise
                    name: str = None,               # name of shape
                    attributes:Attributes = None,   # attributes of shape
        ):

        # init geometry
        Geometry.__init__ (
            self,
            n = n,
            m = m,
            d0 = d0,
            da = da,
            p0 = p0,
            x = x,
            a = a,
            b = b
        )

        # pre calculate parameters
        twist = self.b * height / m
        postRot = twist / 2

        #: Profile of gear
        self.profile = Profile (
                n = n, 
                m = m,
                d0 = d0,
                da = da,
                p0 = p0,
                x = x,
                a = a,
                b = b,
                mhf = mhf,
                rot = postRot + rot,
                innerGear = False,
                name = name,
                attributes = attributes
            )

        # Create shape if no children are supplied
        if children == None:
            children = (self.profile,)

        # Create linear extruded gear wheel
        solid.LinearExtrude.__init__ (self, children, height = height, twist = twist, name = name, attributes = attributes)

        # TODO: clean up rest of attributes
        self.mhf = mhf
        self.rot = rot

        #: Height of wheel.
        self.height = height


    def axisDistance (self, other):
        """
        Cacluclates the axis distance from this wheel to another.
        """
        return self.profile.centerDistance (other)


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            Geometry,
            'children',
            'mhf',
            'rot',
            'name',
            'attributes',
            'height'
        )





class HerringboneWheel (boolean.Union, Geometry):
    """
    Herringbone gear wheel as :ref:`Solid<Solid>`.
    """
    def __init__ (self,
                    n = 20,                         # number of tooth               (de: Zahnanzahl)
                    m = None,                       # module                        (de: Modul)
                    height = 1,                     # height of wheel               (de: Hoehe)
                    d0 = None,                      # base circle diameter          (de: Teilkreisdurchmesser)
                    da = None,                      # addendum circle diameter      (de: Kopfkreisdurchmesser)
                    p0 = None,                      # circular pitch                (de: Umfangsteilung)
                    x = None,                       # profile shift factor          (de: Profilverschiebungsfaktor)
                    a = 20,                         # standard pressure angle       (de: Normaleingriffswinkel)
                    b = 0,                          # helix angle                   (de: Schrägungswinkel)
                    mhf = 0.2,                      # factor of minimum tooth width (de: Faktor für minimale Zahnkopfbreite)
                    children: list = None,          # shape of gearwheel, only for deserialization, will be generated otherwise
                    name: str = None,               # name of shape
                    attributes:Attributes = None,   # attributes of shape
        ):

        assert m != None

        # init geometry
        Geometry.__init__ (
            self,
            n = n,
            m = m,
            d0 = d0,
            da = da,
            p0 = p0,
            x = x,
            a = a,
            b = b
        )

        # pre calculate parameters
        twist = self.b * height / m
        postRot = twist / 2

        #: Profile of gear
        self.profile = Profile (
                n = n, 
                m = m,
                d0 = d0,
                da = da,
                p0 = p0,
                x = x,
                a = a,
                b = b,
                mhf = mhf,
                rot = postRot,
                innerGear = False,
                name = name,
                attributes = attributes
            )

        # Create shape if no children are supplied
        if children == None:
            profile1 = self.profile
            profile2 = Profile (
                n = n, 
                m = m,
                d0 = d0,
                da = da,
                p0 = p0,
                x = x,
                a = a,
                b = -b,
                mhf = mhf,
                rot = -postRot,
                innerGear = False,
                name = name,
                attributes = attributes
            )
            h2 = (height / 2) + constants.Epsilon
            h4 = height / 4
            twist = self.b * height / m
            s1 = transform.Translate (
                solid.LinearExtrude (profile1, h2, twist = twist, name = name, attributes = attributes),
                z = h4
            )
            s2 = transform.Translate (
                solid.LinearExtrude (profile2, h2, twist = -twist, name = name, attributes = attributes),
                z = -h4
            )
            children = (s1, s2)

        # Create linear extruded gear wheel
        boolean.Union.__init__ (self, children, name = name, attributes = attributes)

        # TODO: clean up rest of attributes
        self.mhf = mhf

        #: Height of wheel.
        self.height = height


    def axisDistance (self, other):
        """
        Cacluclates the axis distance from this wheel to another.
        """
        return self.profile.centerDistance (other)


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            Geometry,
            'children',
            'mhf',
            'name',
            'attributes',
            'height'
        )




class InnerWheel (boolean.Difference, Geometry):
    """
    Inner gear wheel as :ref:`Solid<Solid>`.
    """
    def __init__ (self,
                    n = 20,                         # number of tooth               (de: Zahnanzahl)
                    m = None,                       # module                        (de: Modul)
                    height = 1,                     # height of wheel               (de: Hoehe)
                    rimSize = 1,                    # thickness of rim              (de: Randbreite)
                    d0 = None,                      # base circle diameter          (de: Teilkreisdurchmesser)
                    da = None,                      # addendum circle diameter      (de: Kopfkreisdurchmesser)
                    p0 = None,                      # circular pitch                (de: Umfangsteilung)
                    x = None,                       # profile shift factor          (de: Profilverschiebungsfaktor)
                    a = 20,                         # standard pressure angle       (de: Normaleingriffswinkel)
                    b = 0,                          # helix angle                   (de: Schrägungswinkel)
                    mhf = 0.2,                      # factor of minimum tooth width (de: Faktor für minimale Zahnkopfbreite)
                    rot = 0,                        # rotate gear wheel             (de: Rotation des Zahnrads)
                    children: list = None,          # shape of gearwheel, only for deserialization, will be generated otherwise
                    name: str = None,               # name of shape
                    attributes:Attributes = None,   # attributes of shape
        ):

        # init geometry
        Geometry.__init__ (
            self,
            n = n,
            m = m,
            d0 = d0,
            da = da,
            p0 = p0,
            x = x,
            a = a,
            b = b
        )

        # pre calculate parameters
        twist = self.b * height / n
        postRot = twist / 2

        #: Profile of gear
        self.profile = Profile (
                n = n, 
                m = m,
                d0 = d0,
                da = da,
                p0 = p0,
                x = x,
                a = a,
                b = b,
                mhf = mhf,
                rot = postRot + rot,
                innerGear = True,
                name = name,
                attributes = attributes
            )

        # Create shape if no children are supplied
        if children == None:
            outerDiameterMin = self.da + 0.167 * self.m
            outerDiameter = outerDiameterMin + 2 * rimSize

            # outer shape
            cyl = solid.Cylinder (height = height, radius1 = outerDiameter / 2)

            # shape of inner gear wheel
            cutProfile = (self.profile,)

            # inner shape
            inner = solid.LinearExtrude (
                cutProfile,
                height + constants.Epsilon,
                twist
            )

            # create children
            children = (cyl, inner)

        #: Width of rim.
        self.rimSize = rimSize

        # Create linear extruded gear wheel
        boolean.Difference.__init__ (self, children, name = name, attributes = attributes)

        # TODO: clean up rest of attributes
        self.mhf = mhf
        self.rot = rot

        #: Height of wheel.
        self.height = height


    def axisDistance (self, other):
        """
        Cacluclates the axis distance from this wheel to another.
        """
        return self.profile.centerDistance (other)


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            boolean.Difference,
            Geometry,
            'rimSize',
            'mhf',
            'rot',
            'name',
            'attributes',
            'height'
        )




class InnerHerringboneWheel (boolean.Difference, Geometry):
    """
    Inner Herringbone gear wheel as :ref:`Solid<Solid>`.
    """
    def __init__ (self,
                    n = 20,                         # number of tooth               (de: Zahnanzahl)
                    m = None,                       # module                        (de: Modul)
                    height = 1,                     # height of wheel               (de: Hoehe)
                    rimSize = 1,                    # thickness of rim              (de: Randbreite)
                    d0 = None,                      # base circle diameter          (de: Teilkreisdurchmesser)
                    da = None,                      # addendum circle diameter      (de: Kopfkreisdurchmesser)
                    p0 = None,                      # circular pitch                (de: Umfangsteilung)
                    x = None,                       # profile shift factor          (de: Profilverschiebungsfaktor)
                    a = 20,                         # standard pressure angle       (de: Normaleingriffswinkel)
                    b = 0,                          # helix angle                   (de: Schrägungswinkel)
                    mhf = 0.2,                      # factor of minimum tooth width (de: Faktor für minimale Zahnkopfbreite)
                    children: list = None,          # shape of gearwheel, only for deserialization, will be generated otherwise
                    name: str = None,               # name of shape
                    attributes:Attributes = None,   # attributes of shape
        ):

        # init geometry
        Geometry.__init__ (
            self,
            n = n,
            m = m,
            d0 = d0,
            da = da,
            p0 = p0,
            x = x,
            a = a,
            b = b
        )

        # pre calculate parameters
        twist = self.b * height / m
        postRot = twist / 2

        #: Profile of gear
        self.profile = Profile (
                n = n, 
                m = m,
                d0 = d0,
                da = da,
                p0 = p0,
                x = x,
                a = a,
                b = b,
                mhf = mhf,
                rot = postRot,
                innerGear = True,
                name = name,
                attributes = attributes
            )

        # Create shape if no children are supplied
        if children == None:
            # outer shape
            outerDiameterMin = self.da + 0.167 * self.m
            outerDiameter = outerDiameterMin + 2 * rimSize
            cyl = solid.Cylinder (height = height, radius1 = outerDiameter / 2)

            # inner shapes
            profile1 = self.profile
            profile2 = Profile (
                n = n, 
                m = m,
                d0 = d0,
                da = da,
                p0 = p0,
                x = x,
                a = a,
                b = -b,
                mhf = mhf,
                rot = -postRot,
                innerGear = True,
                name = name,
                attributes = attributes
            )
            h2 = (height / 2) + constants.Epsilon
            h4 = height / 4
            twist = self.b * height / m
            s1 = transform.Translate (
                solid.LinearExtrude (profile1, h2, twist = twist, name = name, attributes = attributes),
                z = h4
            )
            s2 = transform.Translate (
                solid.LinearExtrude (profile2, h2, twist = -twist, name = name, attributes = attributes),
                z = -h4
            )
            children = (cyl, boolean.Union((s1, s2)))

        # Create linear extruded gear wheel
        boolean.Difference.__init__ (self, children = children, name = name, attributes = attributes)

        # TODO: clean up rest of attributes
        self.mhf = mhf

        #: Height of wheel.
        self.height = height


    def axisDistance (self, other):
        """
        Cacluclates the axis distance from this wheel to another.
        """
        return self.profile.centerDistance (other)


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            boolean.Difference,
            Geometry,
            'rimSize',
            'mhf',
            'name',
            'attributes',
            'height'
        )



from . import splitLines




class StringWriter:
    """
    Write output stream to a string.
    """
    def __init__(self):
        self.text = ""


    def close(self):
        """
        Close output resource.
        """
        pass # Dummy, not used for StringWriter


    def flush(self):
        """
        Flush buffers.
        """
        pass # Dummy, not used for StringWriter


    def write (self, text_):
        """
        Write text.
        """
        self.text = self.text + str (text_)




class FileWriter:
    """
    Write output stream to a file
    """
    def __init__(self, fileName: str, fileMode: str = "w"):
        self.fileName = fileName
        self.fileMode = fileMode
        self.file = open (fileName, fileMode)


    def close(self):
        if self.file != None:
            self.file.close ()
            self.file = None


    def flush(self):
        if self.file != None:
            self.file.flush ()

    
    def write(self, text_):
        assert self.file != None, "File must not be closed when writing text."
        self.file.write (text_)
    



class TextWriter:
    """
    Used to write structured text to a StreamWriter.
    """
    def __init__(self, writer: object, indentSize: int = 4):
        self.writer = writer
        self.indentSize = indentSize
        self.indent = 0
        self.indentOnLineBegin = None
        self.buffer = ""


    def close (self):
        """
        Closes the StreamWriter behind this TextWriter.
        """
        self.flush ()
        self.writer.close ()
        self.writer = None


    def formatNumber (self, number: float) -> str:
        """
        Formats a number.
        """
        return str (number)


    def writeNumber (self, number: float):
        """
        Write a number.
        """
        return self.write (self.formatNumber (number))


    def write (self, text: str) -> None:
        """
        Write some text.
        """
        if self.indentOnLineBegin == None:
            self.indentOnLineBegin = self.indent
        lines = splitLines (str (text))
        if (len (lines) > 0):
            self.buffer = self.buffer + lines[0]
        if (len (lines) > 1):
            self.flush ()
        for index in range (1, len (lines) - 1):
            self.buffer = lines[index]
            self.flush ()
        if (len (lines) > 1):
            self.buffer = self.buffer + lines[len (lines) - 1]


    def flush (self):
        """
        flushes the current text buffer to the file
        """
        if (self.buffer != ""):
            if self.indentOnLineBegin == None:
                self.indentOnLineBegin = self.indent
            self.writer.write (" " * (self.indentOnLineBegin * self.indentSize))    
            self.writer.write (self.buffer + "\n")
        self.indentOnLineBegin = None
        self.buffer = ""


    def increaseIndent (self) -> None:
        """
        Increase identation used for next write call
        """
        self.indent = self.indent + 1


    def decreaseIndent (self) -> None:
        """
        """
        assert self.indent > 0, "Indentation error"
        self.indent = self.indent - 1


    def __lineEmpty(self):
        """
        Returns true if the current line is empty
        """
        return self.buffer == ""



import math




def CreateMatrix (rows: list):
    """
    Create a matrix.
    """
    assert len (rows) > 0, "matrix must not be empty"
    rl = len (rows[0])
    assert rl > 0, "matrix must not be empty"
    for i in range (1, len (rows)):
        assert rl == len (rows[i]), "all rows must have the same size"
    rowSet = []
    for r in rows:
        rowSet.append (tuple(r))
    return tuple(rowSet)




def multiply (m1, m2):
    """
    Multiply two matrices.
    """
    m1r = len (m1)
    m1c = len (m1[0])       
    m2r = len (m2)
    m2c = len (m2[0])
    rr = m1r
    rc = m2c
    assert m1c == m2r, "Matrix multiplication not defined. Invalid dimensions."
    newRows = []
    for i in range (0, rr):
        newRow = []
        for k in range (0, rc):
            val = 0
            for j in range (0, m1c):
                val = val + (m1[i][j] * m2[j][k])
            newRow.append (val)
        newRows.append (tuple (newRow))
    return tuple(newRows)




def transformPoint2D (matrix, point):
    """
    Special case: transform a 2D point with affinite matrix.
    """
    # TODO: performance optimize
    m1 = (tuple([point[0]]), tuple([point[1]]), tuple([1]))
    tp = multiply (matrix, m1)
    return [tp[0][0], tp[1][0]]




def transformPoint3D (matrix, point):
    """
    Special case: transform a 2D point with affinite matrix.
    """
    # TODO: performance optimize
    m1 = (tuple([point[0]]), tuple([point[1]]), tuple([point[2]]), tuple([1]))
    tp = multiply (matrix, m1)
    return [tp[0][0], tp[1][0], tp[2][0]]




def AffineIdentical2D ():
    return CreateMatrix([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ])




def AffineIdentical3D ():
    return CreateMatrix([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ])




def AffineTranslate2D (x, y):
    return CreateMatrix([
        [1, 0, x],
        [0, 1, y],
        [0, 0, 1]
    ])




def AffineTranslate3D (x, y, z):
    return CreateMatrix([
        [1, 0, 0, x],
        [0, 1, 0, y],
        [0, 0, 1, z],
        [0, 0, 0, 1],
    ])




def AffineScale2D (x, y):
    return CreateMatrix([
        [x, 0, 0],
        [0, y, 0],
        [0, 0, 1]
    ])




def AffineScale3D (x, y, z):
    return CreateMatrix([
        [x, 0, 0, 0],
        [0, y, 0, 0],
        [0, 0, z, 0],
        [0, 0, 0, 1]
    ])




def AffineRotate2D (rz):
    pz = rz * math.pi / 180.0
    return CreateMatrix([
        [math.cos (pz), -math.sin (pz), 0],
        [math.sin (pz), math.cos (pz), 0],
        [0, 0, 1]
    ])




def AffineRotate3Dx (rx):
    px = rx * math.pi / 180.0
    return CreateMatrix([
        [1, 0, 0, 0],
        [0, math.cos (px), math.sin (px), 0],
        [0, -math.sin (px), math.cos (px), 0],
        [0, 0, 0, 1],
    ])




def AffineRotate3Dy (ry):
    py = ry * math.pi / 180.0
    return CreateMatrix([
        [math.cos (py), 0, -math.sin (py), 0],
        [0, 1, 0, 0],
        [math.sin (py), 0, math.cos (py), 0],
        [0, 0, 0, 1],
    ])




def AffineRotate3Dz (rz):
    pz = rz * math.pi / 180.0
    return CreateMatrix([
        [math.cos (pz), -math.sin (pz), 0, 0],
        [math.sin (pz), math.cos (pz), 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ])




def AffineRotate3D (px, py, pz):
    m1 = AffineRotate3Dx (px)
    m2 = AffineRotate3Dy (py)
    m3 = AffineRotate3Dz (pz)
    return multiply (multiply (m3, m2), m1)



import math
import sys
import copy
import os
import re
from . import formatTextBlock
from . import formatTableEntry
from . import formatTableEntryLeftSize




class Option:
    """
    The base class of Flags and Parameters.
    """
    def __init__(self, 
                    name: str,
                    short: str,
                    description: str,
                    default,
                    group,
                    filterFunc
        ):

        #: Name of the Option (required for displaying help).
        self.name = name

        #: Short identifier string (a single char).
        self.short = short

        #: Description of the Option (required for displaying help).
        self.description = description

        #: Group name (required for displaying help).
        self.group = group

        #: Default value for optional Options.
        self.default = default

        #: Value after parsing the commandline.
        self.value = default

        #: Filter function to check parsed values.
        #: def filter (Option, Value) -> str: Function shall return an error string for bad arguments, None if accepted.
        self.filterFunc = filterFunc

        # Internal flag for parser: Option has been parsed
        self.parsed = False




class Flag(Option):
    """
    Flag for command line parser. A flag is simply a boolean flag which is parsed from a command line argument in form **\-\-name** or **\-s(hort)**.
    """
    def __init__(self, name: str, short: str, description: str, group: str, filterFunc = None):
        super().__init__(name, short, description, False, group, filterFunc)


    def onParse(self, parser, value):
        filterResult = None
        if self.filterFunc != None:
            filterResult = self.filterFunc (self, value)
        if filterResult != None:
            return filterResult
        self.parsed = True
        self.value = True
        return None




class Parameter(Option):
    """
    Parameter for command line parser. A parameter is a data value parsed from a command line argument in form **\-\-name value** or **\-s(hort) value**.
    """
    def __init__(self, name: str, short: str, description: str, group: str, default = None, valueName = None, filterFunc = None):
        super().__init__(name, short, description, default, group, filterFunc)

        #: Name of value (required for displaying help).
        self.valueName = valueName


    def onParse(self, parser, value):
        filterResult = None
        if self.filterFunc != None:
            filterResult = self.filterFunc (self, value)
        if filterResult != None:
            return filterResult
        self.parsed = True
        self.value = value
        return None




class ParameterChoose(Parameter):
    """
    Like Parameter, but the choosable options are fixed by a list.
    """
    def __init__(self, name: str, short: str, description: str, group: str, items: dict, default = None, valueName = None, filterFunc = None):
        super().__init__(name, short, description, group, default, valueName, filterFunc)

        #: Name of value (required for displaying help).
        self.valueName = valueName

        #: A dictionary containing the chooseable items. The key is the option string for the command line. Value should contain a dict {'description': aString} for displaying the help.
        self.items = items


    def onParse(self, parser, value):
        if self.filterFunc != None:
            return self.onParse (parser, value)
        if not value in self.items:
            return "has no option \'" + value + "\'"
        self.parsed = True
        self.value = value
        return None




class Argument:
    """
    Argument for command line parser. Arguments get parsed in the order the are registered at the Parser.
    """
    def __init__(self, 
                    name: str, 
                    description: str, 
                    isOptional: bool = False, 
                    optionalParsing: bool = False, 
                    default = None, 
                    isArray: bool = False, 
                    filterFunc = None
                ):

        #: Name of the Argument (required for displaying help).
        self.name = name

        #: Description of the Argument (required for displaying help).
        self.description = description

        #: If True, the Argument is displayed as optional (required for displaying help).
        self.isOptional = isOptional

        #: If True, the Argument is parsed as optional (required for parsing).
        self.optionalParsing = optionalParsing

        #: The Argument is parsed as array and can be used multiple times.
        self.isArray = isArray

        #: Default value for optional Argument.
        self.default = default

        #: Value after parsing the commandline.
        self.value = default

        #: Filter function to check parsed values.
        #: def filter (Option, Value) -> str: Function shall return an error string for bad arguments, None if accepted.
        self.filterFunc = filterFunc

        # Internal flag for parser: Argument has been parsed
        self.parsed = False


    def onParse(self, parser, value):
        filterResult = None
        if self.filterFunc != None:
            filterResult = self.filterFunc (self, value)
        if filterResult != None:
            return filterResult
        self.parsed = True
        self.value = value
        return None




class ArgumentChoose(Argument):
    """
    Like Argument, but the choosable options are fixed by a list.
    """
    def __init__(self, name, description, items: dict, isOptional = False, optionalParsing = False, default = None, isArray = False, filterFunc = None):
        super().__init__(name, description, isOptional = isOptional, optionalParsing = optionalParsing, default = default, isArray = isArray, filterFunc = filterFunc)
        self.items = items


    def onParse(self, parser, value):
        if self.filterFunc != None:
            return self.onParse (parser, value)
        if not value in self.items:
            return "has no option \'" + value + "\'"
        self.parsed = True
        self.value = value
        return None




class ExtraArguments:
    """
    ExtraArguments takes all extra command line arguments after **\-\-** was parsed. Arguments **\-\-** are not interpreted by the Parser.
    """
    def __init__(self, name, description):
        self.name = name
        self.description = description
        self.value = []
        self.parsed = False


    def onParse(self, parser, value):
        self.parsed = True
        self.value = value
        return None




def FilterNumber(param, value):
    """
    Filter: Assert number.
    """
    if re.match(r'^-?\d+(?:\.\d+)?$', value) is None:
        return "parameter --" + param.name + " expectes a number as value"
    return None




def FilterNumberGreterZero(param, value):
    """
    Filter: Assert number greater than zero.
    """
    if re.match(r'^-?\d+(?:\.\d+)?$', value) is None:
        return "parameter --" + param.name + " expectes a number as value"
    if float (value) <= 0:
        return "parameter --" + param.name + " expectes a number greater zero"
    return None




def FilterInteger(param, value):
    """
    Filter: Assert integral number.
    """
    if value.isnumeric () == False:
        return "parameter --" + param.name + " expectes an integer as value"
    return None




def FilterIntegerGreterZero(param, value):
    """
    Filter: Assert integral number greater than zero.
    """
    if value.isnumeric () == False:
        return "parameter --" + param.name + " expectes an integer as value"
    if int (value) <= 0:
        return "parameter --" + param.name + " expectes an integer greater zero"
    return None




class Parser:
    """
    Parser for command line with integrated help system.
    """
    def __init__(self, title = None, description = None):
        self.title = title
        self.description = description
        self.entries = []
        self.indent = 2
        self.centerSpace = self.indent
        # TODO: get line size on runtime
        self.defaultLineSize = 80
        

    def register(self, entry):
        """
        Register flag, parameter or argument.
        """
        hasArray = False
        hasOptional = False
        hasExtraArgs = False
        for e in self.entries:
            if isinstance (e, ExtraArguments):
                hasExtraArgs = True
            if isinstance (e, Argument):
                if e.isOptional:
                    hasOptional = True
                if e.isArray:
                    hasArray = True
        if isinstance (entry, (Flag, Parameter)):
            for e in self.entries:
                if isinstance (e, (Flag, Parameter, ParameterChoose)):
                    if e.name == entry.name:
                        assert False, "an option with name --" + str (e.name) + " already exists." 
                    if ((entry.short != None) and (e.short == entry.short)):
                        assert False, "an option with name -" + str (e.short) + " already exists." 
            self.entries.append (entry)
        elif isinstance (entry, Argument):
            if hasArray:
                assert False, "can not add an argument after an array of arguments"
            if hasOptional:
                if entry.isOptional == False:
                    assert False, "can not add a non-optional argument after an optional argument"
            self.entries.append (entry)
        elif isinstance (entry, ExtraArguments):
            assert hasExtraArgs == False, "can not add ExtraArguments twice."
            self.entries.append (entry)
        else:
            assert False, "type of entry is invalid"


    def showHelp(self, withDescription = True):
        """
        Show command line help.
        """
        # Short title
        appName = ""
        if len (sys.argv) > 0:
            appName = str (sys.argv[0])
        if self.title != None:
            print (appName + " - " + self.title)
        else:
            print (appName)
        
        # Usage
        print ("\nusage: " + appName + " " + self._formatUsage ())

        # Description
        if withDescription and (self.description != None):
            print ("\n" + self.description)

        # Calculate table sizes
        lineSize = self.defaultLineSize
        minLineSize = 40
        if lineSize < minLineSize:
            lineSize = minLineSize

        # Table formatting
        indent = self.indent
        centerSpace = self.centerSpace - 1
        if centerSpace < 0:
            centerSpace = 0
        leftColSize = math.floor (lineSize * 0.4)
        minLineSize = leftColSize + centerSpace + 10 + indent
        if lineSize < minLineSize:
            lineSize = minLineSize

        # determine size of left column
        minLeftSize = 0
        for entry in self.entries:
            if isinstance (entry, Argument):
                if (len (entry.name) > minLeftSize):
                    minLeftSize = len (entry.name)
        for entry in self.entries:
            if isinstance (entry, (Flag, Parameter)):
                l, r = self._formatOption (entry, indent, 1000)
                ls = formatTableEntryLeftSize (l, lineSize, leftColSize, centerSpace, indent = indent) + indent
                if (ls > minLeftSize):
                    minLeftSize = ls
        minLeftSize = minLeftSize
        if minLeftSize < leftColSize:
            leftColSize = minLeftSize
        if leftColSize < 10:
            leftColSize = 10

        leftColSize = leftColSize - indent

        # maximum right table size
        maxRightSize = lineSize - centerSpace - leftColSize - indent

        # Command line arguments
        first = True
        for entry in self.entries:
            if isinstance (entry, Argument):
                if first:
                    print ("\nArguments:")
                    first = False
                l = entry.name
                r = entry.description
                if isinstance (entry, ArgumentChoose):
                    astr = self._formatChooseables (entry, "", indent, maxRightSize)
                    if astr != "":
                        astr = "\n" + astr
                    r = r + astr
                if entry.default != None:
                    if r != "":
                        r = r + "\n"
                    r = r + "Default value: " + str (entry.default)
                s = formatTableEntry (l, r, lineSize, leftColSize, centerSpace, indent = indent, prefix = " " * indent)
                print (s)
        for entry in self.entries:
            if isinstance (entry, ExtraArguments):
                if first:
                    print ("\nArguments:")
                    first = False
                l = "-- " + entry.name
                r = entry.description
                s = formatTableEntry (l, r, lineSize, leftColSize, centerSpace, indent = indent, prefix = " " * indent)
                print (s)

        # Command line flags and options
        groups = {}
        groupOrder = []
        for entry in self.entries:
            if isinstance (entry, (Flag, Parameter)):
                if entry.group in groups:
                    groups[entry.group].append (entry)
                else:
                    groups[entry.group] = [entry]
                    groupOrder.append (entry.group)
        for groupName in groupOrder:
            first = True
            for entry in groups[groupName]:
                if first:
                    print ("\n" + groupName + ":")
                    first = False
                l, r = self._formatOption (entry, indent, maxRightSize)
                s = formatTableEntry (l, r, lineSize, leftColSize, centerSpace, indent = indent, prefix = " " * indent)
                print (s)


    def _formatUsage(self):
        """
        Format usage for displaying help.
        """
        text = ""
        for entry in self.entries:
            if isinstance (entry, (Parameter, Flag)):
                text = "[--<options> ...]"
                break
        for entry in self.entries:
            if isinstance (entry, Argument):
                if text != "":
                    text = text + " "
                en = entry.name
                if entry.isArray:
                    en = en + " ..."
                if entry.isOptional:
                    en = "[" + en + "]"
                text = text + en
        for entry in self.entries:
            if isinstance (entry, ExtraArguments):
                if text != "":
                    text = text + " "
                text = text + " [-- " + entry.name + " ...]"
        return text


    def _formatOption(self, option, indent, maxRightSize):
        """
        Format an option for displaying help.
        """
        text = "--" + option.name
        if option.short != None:
            if option.short != "":
                text = text + ", -" + option.short
        if isinstance(option, Parameter):
            text = text + " $" + option.valueName
        desc = ""
        if option.description != None:
            desc = option.description
        desc = self._formatChooseables (option, desc, indent, maxRightSize)
        if isinstance (option, Parameter):
            if option.default != None:
                if desc != None:
                    desc = desc + "\n"
                desc = desc + "Default value: " + str (option.default)
        return (text, desc)


    def _formatChooseables(self, entry, oldText, indent, maxRightSize):
        """
        Format choosable options.
        """
        newText = oldText
        if not isinstance (entry, (ParameterChoose, ArgumentChoose)):
            return oldText
        emitNewLine = False
        if newText != "":
            emitNewLine = True
        for key in entry.items:
            ostr = "\'" + key + "\'"
            mi = entry.items[key]
            if "description" in mi:
                descr = mi["description"]
                if descr != None:
                    if descr != "":
                        idescr = formatTextBlock (descr, indent, maxRightSize - indent)
                        ostr = ostr + ":\n" + idescr
            if emitNewLine:
                newText = newText + "\n"    
            newText = newText + formatTextBlock (ostr, indent, maxRightSize - indent)
            emitNewLine = True
        return newText


    def commandlineInvalid(self, message):
        """
        Print error and terminate application with error code 1.
        """
        appName = ""
        if len (sys.argv) > 0:
            appName = str (sys.argv[0])
        print ("error parsing command line: " + message)
        print ("run \'" + appName + " --help\' for more informations.")
        os._exit (1)


    def parse(self, arguments = None):
        """
        Parse command line arguments.
        """
        # Get arguments
        if arguments != None:
            args = copy.copy (arguments)
        else:
            args = copy.copy (sys.argv)
            args.pop (0)

        # parse flags and parameters and extra arguments
        newArgs = []
        argIndex = 0
        while argIndex < len (args):
            if args[argIndex] == "--":
                # check for extra arguemnts entries
                extraEntry = None
                for p in self.entries:
                    if isinstance (p, ExtraArguments):
                        extraEntry = p
                        break
                if extraEntry == None:
                    self.commandlineInvalid ("missing option name after --")            
                else:
                    argIndex = argIndex + 1
                    additionalArgs = []
                    while argIndex < len (args):
                        additionalArgs.append (args[argIndex])
                        argIndex = argIndex + 1 
                    extraEntry.onParse (self, additionalArgs)
                    break

            elif args[argIndex].startswith ("--"):
                # check for long option entry
                argName = args[argIndex][2 : len (args[argIndex])]
                argIndex = argIndex + 1
                argFound = False
                for p in self.entries:
                    if isinstance(p, (Flag, Parameter)):
                        if p.name == argName:
                            argFound = True
                            if isinstance (p, Parameter):
                                if argIndex >= len (args):
                                    self.commandlineInvalid ("missing value after --" + argName)            
                                argValue = args[argIndex]
                                argIndex = argIndex + 1
                                pe = p.onParse (self, argValue)
                                if pe != None:
                                    self.commandlineInvalid ("parameter --" + argName + " " + pe)
                            else:
                                pe = p.onParse (self, True)
                                if pe != None:
                                    self.commandlineInvalid ("parameter --" + argName + " " + pe)
                            break
                if argFound == False:
                    self.commandlineInvalid ("option --" + argName + " is unknown")

            elif args[argIndex].startswith ("-"):
                # check for short option entry
                tok = args[argIndex][1 : len (args[argIndex])]
                if len (tok) < 1:
                    self.commandlineInvalid ("expected option after -")
                while len (tok) > 0:
                    opt = tok[0]
                    argFound = False
                    for p in self.entries:
                        if isinstance(p, (Flag, Parameter)):
                            if p.short == opt:
                                argFound = True
                                if isinstance (p, Parameter):
                                    if len (tok) > 1:
                                        argValue = tok[1 : len (tok)]
                                        tok = ""
                                    else:
                                        argIndex = argIndex + 1
                                        if argIndex >= len (args):
                                            self.commandlineInvalid ("missing value after -" + opt)
                                        argValue = args[argIndex]
                                    pe = p.onParse (self, argValue)
                                    if pe != None:
                                        self.commandlineInvalid ("parameter -" + opt + " " + pe)
                                else:
                                    pe = p.onParse (self, True)
                                    if pe != None:
                                        self.commandlineInvalid ("parameter -" + opt + " " + pe)
                                break
                    if argFound == False:
                        self.commandlineInvalid ("option -" + opt + " is unknown")
                    if len (tok) > 0:
                        tok = tok[1 : len (tok)]
                argIndex = argIndex + 1

            else:
                # pass to argument parsing
                newArgs.append (args[argIndex])
                argIndex = argIndex + 1
        args = newArgs

        # parse arguments
        for entry in self.entries:
            if isinstance (entry, Argument):
                if len (args) < 1:
                    break
                if entry.isArray:
                    pe = entry.onParse (self, args)
                    if pe != None:
                        self.commandlineInvalid ("argument " + entry.name + ": " + pe)
                    args = []
                    break
                else:
                    pe = entry.onParse (self, args[0])
                    if pe != None:
                        self.commandlineInvalid ("argument " + entry.name + ": " + pe)
                    args.pop (0)


        # assert no arguments left
        if len (args) > 0:
            self.commandlineInvalid ("invalid extra arguments: \'" + str (args[0]) + "\'")

        # assert all parameters are parsed
        for entry in self.entries:
            if isinstance (entry, Argument):
                if (entry.parsed == False) and (entry.optionalParsing == False) and (entry.isOptional == False):
                    self.commandlineInvalid ("missing argument " + str (entry.name))
    

    def dumpParsed(self, withDefaults = False, prefix = "  "):
        """
        Dump parsed command line options.
        """
        for entry in self.entries:
            if isinstance (entry, Parameter):
                if withDefaults or entry.parsed:
                    print (prefix + "--" + entry.name + " = " + str (entry.value))
            if isinstance (entry, Flag):
                if entry.parsed:
                    print (prefix + "--" + entry.name)
        for entry in self.entries:
            if isinstance (entry, Argument):
                if entry.parsed or withDefaults:
                    print (prefix + entry.name + " = " + str (entry.value))
        for entry in self.entries:
            if isinstance (entry, ExtraArguments):
                if entry.parsed or withDefaults:
                    print (prefix + "-- " + str (entry.value))



import subprocess
import os
import math
import threading
from ... import tree
from ... import solid
from ... import shape
from ... import scene
from ... import transform
from ... import boolean
from ... import material
from .. import textwriter
from .. import downsample
from .. import cmath
from .. import matrix
from ...algorithms import reducegroups




# WARNING: PovRay swaps X/Z dimensions compared to OpenScad - this is the reason, we are swapping it back here.
# Also rotations are in reverse direction.




# Can be replaced with an absolute path to the povray executable
PovRayExecutable = "povray"




# Constant for generating distances
LargeEpsilon = 0.1




# TODO: move bezier type to other class
class PovSolidPrism(tree.Node):
    """
    Prism rendering primitive for PovRay.
    """
    # TODO: smoothing
    def __init__(self,
                    polygon,
                    height,
                    centerOffset = 0,
                    name: str = "",
                    attributes: dict = None
        ):
        super().__init__(name = name, attributes = attributes)
        self.polygon = polygon
        self.height = height
        self.centerOffset = centerOffset

        if isinstance (polygon, shape.Polygon):
            type_ = "linear_spline"
        elif isinstance (polygon, shape.QuadraticSpline):
            type_ = "quadratic_spline"
        elif isinstance (polygon, shape.CubicSpline):
            type_ = "cubic_spline"
        elif isinstance (polygon, shape.BezierSpline):
            type_ = "bezier_spline"
        else:
            assert False, "unexpected shape type"
        
        #: interpolation type
        self.type_ = type_



class _PovRayLinearExtrudeRewriter(tree.Visitor):
    """
    Iterate a tree, find all linear extrusions and replace them...
    """
    def __init__(self):
        super().__init__()
        self._rule (tree.Generator, self.enterGenerator)
        self._rule (solid.LinearExtrude, self.enterExtrusion, self.leaveExtrusion)
        self._rule (shape.Circle, self.enterCircle)
        self._rule (shape.Square, self.enterSquare)
        self._rule (shape.Polygon, self.enterPolygon)
        self._rule (shape.QuadraticSpline, self.enterPolygon)
        self._rule (shape.CubicSpline, self.enterPolygon)
        self._rule (shape.BezierSpline, self.enterPolygon)
        self._rule (shape.PolarShape, self.enterPolarShape, self.leaveTransformedShape)
        self._rule (transform.Translate, self.enterTransform, self.leaveTransform)
        self._rule (transform.Rotate, self.enterTransform, self.leaveTransform)
        self._rule (transform.Scale, self.enterTransform, self.leaveTransform)


    def execute(self, node, defaultAttributes):
        ctx = tree.VisitorContext (defaultAttributes)
        ctx.extrusionData = None
        return self.visit (node, ctx)


    def _enterDefault(self, node, ctx):
        """
        Enter default node type.
        """
        assert (ctx.extrusionData == None), "found unhandled type in extrusion"
        return True


    def enterGenerator(self, node, ctx):
        self.visit (node.generate (ctx.getResolvedAttributes ()), ctx)


    def enterCircle(self, node, ctx):
        """
        Enter a circle.
        """
        if ctx.extrusionData == None:
            return True

        height=ctx.extrusionData["extrudeRoot"].evaluate ("height", ctx.getResolvedAttributes ())
        radius=node.evaluate ("radius", ctx.getResolvedAttributes ())
        mergedAttributs=node.attributes
        ctx.extrusionData["generated"] = self.applyTransformStack (solid.Cylinder (height=height, radius=radius, attributes=mergedAttributs), ctx)
        return False


    def enterSquare(self, node, ctx):
        """
        Enter a square.
        """
        if ctx.extrusionData == None:
            return True

        height=ctx.extrusionData["extrudeRoot"].evaluate ("height", ctx.getResolvedAttributes ())
        size=node.evaluate ("size", ctx.getResolvedAttributes ())
        mergedAttributs=node.attributes
        ctx.extrusionData["generated"] = self.applyTransformStack (solid.Cube ([size[0], size[1], height], attributes=mergedAttributs), ctx)
        return False


    def enterPolygon(self, node, ctx):
        """
        Enter a polygon.
        """
        if ctx.extrusionData == None:
            return True
        mergedAttributs=node.attributes
        height=ctx.extrusionData["extrudeRoot"].evaluate ("height", ctx.getResolvedAttributes ())
        ctx.extrusionData["generated"] = self.applyTransformStack (PovSolidPrism (node, height, attributes=mergedAttributs), ctx)
        return False


    # TODO: Need this? Generate class should do the magic here.
    def enterPolarShape(self, node, ctx):
        """
        Enter a polar shape, transform to a simple shape and re-run on visitor (self)
        """
        transformed = node.toQuadraticSpline (ctx.getResolvedAttributes ())
        if transformed == None:
            transformed = node.toCubicSpline (ctx.getResolvedAttributes ())
        if transformed == None:
            transformed = node.toPolygon (ctx.getResolvedAttributes ())
        assert transformed != None, "can not transform to a simple shape geometry"
        self.visit (transformed, ctx)
        return False


    def leaveTransformedShape(self, node, ctx):
        """
        Leave transformed shape.
        """
        return node


    def applyTransformStack(self, node, ctx):
        """
        Apply transform stack to newly created node (2D -> 3D)
        """
        transformStack = ctx.extrusionData["transformStack"]
        newNode = node
        for index in range (len (transformStack) - 1, -1, -1):
            t = transformStack[index]
            if isinstance (t, transform.Translate):
                tv = t.evaluate ("vector", ctx.getResolvedAttributes ())
                x = tv[0]
                y = tv[1]
                newNode = transform.Translate (newNode, [x, y, 0])
            elif isinstance (t, transform.Rotate):
                tv = t.evaluate ("vector", ctx.getResolvedAttributes ())
                x = tv[0]
                newNode = transform.Rotate (newNode, [0, 0, x])
            elif isinstance (t, transform.Scale):
                tv = t.evaluate ("vector", ctx.getResolvedAttributes ())
                x = tv[0]
                y = tv[1]
                newNode = transform.Scale (newNode, [x, y, 1])
            else:
                assert False, "transformation type not handled."
        return newNode


    def enterTransform(self, node, ctx):
        """
        Enter a transformation.
        """
        if ctx.extrusionData == None:
            return True

        ctx.extrusionData["transformStack"].append (node)
        return True


    def leaveTransform(self, node, ctx):
        """
        Leave a transformation.
        """
        return node


    def enterExtrusion(self, node, ctx):
        """
        Enter extrusion.
        """
        assert ctx.extrusionData == None, "found extrusion in extrusion."
        twist = node.parameter ("twist")
        if twist in (None, 0):
            ctx.extrusionData = {
                "extrudeRoot": node,
                "transformStack": [],
                "generated": None
            }
            return True
        else:
            attrs = ctx.getResolvedAttributes ()
            poly = node.asPolyhedron (attrs)
            assert poly != None, "can not convert twisted linear extrusion in polyhedron"
            # TODO: smoothing
            smoothing = False if not 'smoothing' in attrs else attrs['smoothing']
            if smoothing:
                smoothingAngle = tree.DefaultAttributes['smoothing.thresholdAngle'] if not 'smoothing.thresholdAngle' in attrs else attrs['smoothing.thresholdAngle']
                poly = poly.optimize ().smoothEdgeNormals (angleThreshold = smoothingAngle)
            ctx.extrusionData = {
                "extrudeRoot": node,
                "transformStack": [],
                "generated": poly
            }
            return False


    def leaveExtrusion(self, node, ctx):
        """
        Leave extrusion.
        """
        newItem = ctx.extrusionData["generated"]
        ctx.extrusionData = None
        return newItem

PovRayLinearExtrudeRewriter = _PovRayLinearExtrudeRewriter()






class PovSolidExtrudedTorus(tree.Tree):
    """
    Torus rendering primitive for PovRay.
    """
    # TODO: smoothing
    def __init__(self,
                    points,
                    angle,
                    transform,
                    type_: str = "linear_spline",
                    name: str = "",
                    attributes: dict = None
        ):
        super().__init__(name = name, attributes = attributes)
        self.points = points
        self.angle = angle
        self.transform = transform
        self.type_ = type_
        self.applyTransform ()
        self.calculateBoundingBox ()

    
    def applyTransform(self):
        """
        Applies the affinite transform to all control points.
        """
        for i in range(0, len (self.points)):
            self.points[i] = matrix.transformPoint2D (self.transform, self.points[i])


    def calculateBoundingBox(self):
        """
        Calculates the bounding of a rotate extrude object.
        """
        # calculate bounding box as cylinder around rotated shape.
        radius = 0
        heightMin = None
        heightMax = None
        for p in self.points:
            if p[0] < 0:
                if radius < -p[0]:
                    radius = -p[0]
            else:
                if radius < p[0]:
                    radius = p[0]
            if heightMin == None:
                heightMin = p[1]
            elif heightMin > p[1]:
                heightMin = p[1]
            if heightMax == None:
                heightMax = p[1]
            elif heightMax < p[1]:
                heightMax = p[1]
        self.bbHeightMin = heightMin if heightMin != None else 0
        self.bbHeightMax = heightMax if heightMax != None else 0
        self.bbRadius = radius





# TODO: move to cmath
def warpAngle(angle):
    """
    Return angle in range 0..360
    """
    # TODO: optimize, remove loop on negative angles
    while (angle < 0):
        angle = angle + 360.0
    return angle % 360.0



def createRotateAngleLimiter(rotShape, se, sa = None, inverse = False):
    """
    Create angle limiter for rotation extrude. 
    """
    # calculate angles and apply defaults
    if sa == None:
        sa = 0

    # clamp angles and swap if required
    se = warpAngle (se)
    sa = warpAngle (sa)
    if se < sa:
        t = sa
        sa = se
        se = t

    # get geometry of bouding box
    sr = math.sqrt (rotShape.bbRadius * rotShape.bbRadius) + LargeEpsilon
    h1 = rotShape.bbHeightMin
    h2 = rotShape.bbHeightMax

    # quadrant control points (counter clock wise)
    qccw = [
            [sr, 0],
            [sr, sr],
            [0, sr],
            [-sr, sr],
            [-sr, 0],
            [-sr, -sr],
            [0, -sr],
            [sr, -sr],
            [sr, 0],
            [sr, 0]
        ]

    # quadrant control points (clock wise)
    qcw = [
            [sr, 0],
            [sr, -sr],
            [0, -sr],
            [-sr, -sr],
            [-sr, 0],
            [-sr, sr],
            [0, sr],
            [sr, sr],
            [sr, 0],
            [sr, 0]
        ]

    # select quadrant control points
    q = qccw if inverse == False else qcw

    # calculate control point on quadrant
    cp = lambda angle, cp1, cp2: [
            (cp2[0] * math.tan (angle * math.pi / 180)) + (cp1[0] * (1 - math.tan (angle * math.pi / 180))),
            (cp2[1] * math.tan (angle * math.pi / 180)) + (cp1[1] * (1 - math.tan (angle * math.pi / 180)))
        ]

    # get quadrant of first sa, se
    saq = int (math.floor (sa / 45.0))
    seq = int (math.floor (se / 45.0))
    sar = sa - (saq * 45)
    ser = se - (seq * 45)

    # calculate points for polygon
    points = [[0, 0]]
    points.append (cp (sar, q[saq], q[saq + 1]))
    for qi in range (saq, seq):
        points.append (q[qi + 1])
    points.append (cp (ser, q[seq], q[seq + 1]))

    # calculate prism
    h = (2 * LargeEpsilon) + h2 - h1
    ch = (h1 + h2) / 2
    return PovSolidPrism (shape.Polygon (points), h, centerOffset = ch)





class _PovRayRotateExtrudeRewriter(tree.Visitor):
    """
    Iterate a tree, find all rotatory extrusions and replace them...
    """
    #special case: normal torus if circle and not scaled / rotated (decide rules)
    def __init__(self):
        super().__init__()
        self._rule (solid.RotateExtrude, self.enterExtrusion, self.leaveExtrusion)
        self._rule (shape.Circle, self.enterCircle)
        self._rule (shape.Square, self.enterSquare)
        self._rule (shape.Polygon, self.enterPolygon)
        self._rule (shape.QuadraticSpline, self.enterPolygon)
        self._rule (shape.CubicSpline, self.enterPolygon)
        self._rule (shape.BezierSpline, self.enterPolygon)
        self._rule (shape.PolarShape, self.enterPolarShape, self.leaveTransformedShape)
        self._rule (transform.Translate, self.enterTransform, self.leaveTransform)
        self._rule (transform.Rotate, self.enterTransform, self.leaveTransform)
        self._rule (transform.Scale, self.enterTransform, self.leaveTransform)


    def execute(self, node, defaultAttributes):
        ctx = tree.VisitorContext (defaultAttributes)
        ctx.extrusionData = None
        return self.visit (node, ctx)


    def _enterDefault(self, node, ctx):
        """
        Enter default node type.
        """
        assert (ctx.extrusionData == None), "found unhandled type in extrusion"
        return True


    def enterSquare(self, node, ctx):
        """
        Enter a square.
        """
        if ctx.extrusionData == None:
            return True

        angle=ctx.extrusionData["extrudeRoot"].evaluate ("angle", ctx.getResolvedAttributes ())
        size=node.evaluate ("size", ctx.getResolvedAttributes ())
        mergedAttributs=node.attributes
        sx = size[0] / 2.0
        sy = size[1] / 2.0
        points = [
            [-sx, -sy],
            [sx, -sy],
            [sx, sy],
            [-sx, sy]
        ]
        transform = self.createTransformFromStack (ctx)
        newItem = PovSolidExtrudedTorus (points, angle, transform, attributes=mergedAttributs)
        ctx.extrusionData["generated"] = self.limitExtrusionAngle (newItem, angle, ctx)
        return False


    def enterCircle(self, node, ctx):
        """
        Enter a circle.
        """
        # control point for center of quadrant of circle        
        c = 0.55191502449

        # split cirlcle into 4 quadrants of qubic bezier curves
        qbPoints = [
            [0, 1],
            [c, 1],
            [1, c],
            [1, 0],
            [1, 0],
            [1, -c],
            [c, -1],
            [0, -1],
            [0, -1],
            [-c, -1],
            [-1 , -c],
            [-1, 0],
            [-1, 0],
            [-1, c],
            [-c, 1],
            [0, 1]
        ]

        # apply own size as scale to transform stack
        radius=node.evaluate ("radius", ctx.getResolvedAttributes ())
        ctx.extrusionData["transformStack"].append (transform.Scale (vector = [radius, radius]))
        gt = self.createTransformFromStack (ctx)
        ctx.extrusionData["transformStack"].pop ()
        angle=ctx.extrusionData["extrudeRoot"].evaluate ("angle", ctx.getResolvedAttributes ())
        mergedAttributs=node.attributes
        newItem = PovSolidExtrudedTorus (qbPoints, angle, gt, type = "bezier_spline", attributes=mergedAttributs)
        ctx.extrusionData["generated"] = self.limitExtrusionAngle (newItem, angle, ctx)
        return False


    def enterPolygon(self, node, ctx):
        """
        Enter a polygon.
        """
        if ctx.extrusionData == None:
            return True
        angle=ctx.extrusionData["extrudeRoot"].evaluate ("angle", ctx.getResolvedAttributes ())
        points=node.getData ()
        mergedAttributs=node.attributes
        transform = self.createTransformFromStack (ctx)
        if isinstance (node, shape.QuadraticSpline):
            type_ = "quadratic_spline"
        elif isinstance (node, shape.CubicSpline):
            type_ = "cubic_spline"
        elif isinstance (node, shape.BezierSpline):
            type_ = "bezier_spline"
        elif isinstance (node, shape.Polygon):
            type_ = "linear_spline"
        else:
            assert False, "unexpected shape type"
        newItem = PovSolidExtrudedTorus (points, angle, transform, attributes=mergedAttributs, type_=type_)
        ctx.extrusionData["generated"] = self.limitExtrusionAngle (newItem, angle, ctx)
        return False


    def enterPolarShape(self, node, ctx):
        """
        Enter a polar shape, transform to a simple shape and re-run on visitor (self)
        """
        transformed = node.toQuadraticSpline (ctx.getResolvedAttributes ())
        if transformed == None:
            transformed = node.toCubicSpline (ctx.getResolvedAttributes ())
        if transformed == None:
            transformed = node.toPolygon (ctx.getResolvedAttributes ())
        assert transformed != None, "can not transform to a simple shape geometry"
        self.visit (transformed, ctx)
        return False


    def leaveTransformedShape(self, node, ctx):
        """
        Leave transformed shape.
        """
        return node


    def limitExtrusionAngle(self, node, angle, ctx):
        """
        Limit angle of rotate extrude.
        """
        if angle < 0:
            if angle > -360:
                rotLimit = createRotateAngleLimiter (node, 0, -angle, inverse = True).intersect (node)
                return rotLimit
        elif angle < 360:
            rotLimit = createRotateAngleLimiter (node, 0, angle).intersect (node)
            return rotLimit
        return node


    def createTransformFromStack(self, ctx):
        """
        Create affinite transform from transform stack
        """
        transformStack = ctx.extrusionData["transformStack"]
        m = matrix.AffineIdentical2D ()
        for index in range (len (transformStack) - 1, -1, -1):
            se = transformStack[index]
            nt = None
            if isinstance (se, transform.Translate):
                tv = se.evaluate ("vector", ctx.getResolvedAttributes ())
                x = tv[0]
                y = tv[1]
                nt = matrix.AffineTranslate2D (x, y)
            elif isinstance (se, transform.Scale):
                tv = se.evaluate ("vector", ctx.getResolvedAttributes ())
                x = tv[0]
                y = tv[1]
                nt = matrix.AffineScale2D (x, y)
            elif isinstance (se, transform.Rotate):
                tv = se.evaluate ("vector", ctx.getResolvedAttributes ())
                x = tv[0]
                nt = matrix.AffineRotate2D (x)
            else:
                assert False, "transformation type not handled."
            m = matrix.multiply (nt, m)
        return m


    def enterTransform(self, node, ctx):
        """
        Enter a transformation.
        """
        if ctx.extrusionData == None:
            return True

        ctx.extrusionData["transformStack"].append (node)
        return True


    def leaveTransform(self, node, ctx):
        """
        Leave a transformation.
        """
        return node


    def enterExtrusion(self, node, ctx):
        """
        Enter extrusion.
        """
        assert ctx.extrusionData == None, "found extrusion in extrusion."
        angle = node.parameter ("angle")
        # TODO: move assert to tree node
        assert (angle >= -360) and (angle <= 360)
        ctx.extrusionData = {
            "extrudeRoot": node,
            "transformStack": [],
            "generated": None
        }
        return True


    def leaveExtrusion(self, node, ctx):
        """
        Leave extrusion.
        """
        newItem = ctx.extrusionData["generated"]
        ctx.extrusionData = None
        return newItem

PovRayRotateExtrudeRewriter = _PovRayRotateExtrudeRewriter ()





class _PovRaySceneWriter (tree.Visitor):
    """
    Write scene items like lights to PovRay script.
    """
    # TODO: also allow translations on lights
    def __init__(self):
        super().__init__()
        self._rule (scene.Scene, self.unhandledSceneElement)
        self._rule (scene._Light, self.enterSkip)
        self._rule (scene.PovCameraRefinement, self.enterSkip)
        self._rule (scene.PhotonSettings, self.enterSkip)
        self._rule (scene.RadiositySettings, self.enterSkip)


    def enterSkip(self, node, ctx):
        """
        Skip scene item.
        """
        return True


    def unhandledSceneElement(self, node, ctx):
        """
        Enter a un handled scene element.
        """
        assert False, "scene element not handled"
        return True
        

    def execute(self, node, parentContext):
        """
        Execute scene writer.
        """
        ctx = tree.VisitorContext ()
        ctx.parentContext = parentContext
        ctx.writer = parentContext.writer
        self.visit (node, ctx)


PovRaySceneWriter = _PovRaySceneWriter ()





class _PovRayIncludeCollector (tree.Visitor):
    """
    Collect include files of for povray
    """
    def __init__(self):
        super().__init__ ()
        self._rule (solid.Solid, self.enterMaterial)
        self._rule (scene._Light, self.enterMaterial)


    def execute(self, node, defaultAttributes):
        """
        Run collection of include files
        """
        ctx = tree.VisitorContext ()
        ctx.includeFiles = []
        self.visit (node, ctx)
        return ctx.includeFiles


    def enterMaterial(self, node, ctx):
        """
        Collect include files from material parameters.
        """
        includes = ctx.getAttribute ("material.povincludes")
        if includes != None:
            if isinstance (includes, list):
                for include in includes:
                    if not include in ctx.includeFiles:
                        ctx.includeFiles.append (include)
            if isinstance (includes, str):
                if not includes in ctx.includeFiles:
                    ctx.includeFiles.append (includes)
        return True


PovRayIncludeCollector = _PovRayIncludeCollector ()





class _PovRayCameraRefinementCollector (tree.Visitor):
    """
    Collect include files of for povray
    """
    def __init__(self):
        super().__init__ ()
        self._rule (scene.PovCameraRefinement, self.enterRefinement)


    def execute(self, node, defaultAttributes):
        """
        Run collection of include files
        """
        ctx = tree.VisitorContext ()
        ctx.refinements = ""
        self.visit (node, ctx)
        return ctx.refinements


    def enterRefinement(self, node, ctx):
        """
        Collect camera refinement parameters.
        """
        ctx.refinements = ctx.refinements + node.format () + "\n"
        return True


PovRayCameraRefinementCollector = _PovRayCameraRefinementCollector ()





class PovRayTransformStack:
    """
    Transform stack for PovRay writer.
    """
    def __init__(self):
        self.stacks = [[]]


    def pushLayer(self):
        """
        Push a new transform layer to the stack.
        """
        self.stacks.append ([])


    def popAndWriteLayer(self, ctx):
        """
        Pops a transform layer from the stack and writes it.
        """
        assert len (self.stacks) > 1
        self.stacks.pop ()
        stack = self.stacks[-1]
        for ti in range (len (stack) - 1, -1, -1):
            t = stack[ti]
            self.writeTransform(t, ctx)


    def writeTransform(self, node, ctx):
        """
        Writes a single transform to the povray script.
        """
        if isinstance (node, transform.Translate):
            tv = node.evaluate ("vector", ctx.getResolvedAttributes ())
            ctx.writer.write ("translate <" + str (tv[0]) + ", " + str (tv[2]) + ", " + str (tv[1]) + ">\n")
        elif isinstance (node, transform.Rotate):
            tv = node.evaluate ("vector", ctx.getResolvedAttributes ())
            ctx.writer.write ("rotate <" + str (-tv[0]) + ", " + str (-tv[2]) + ", " + str (-tv[1]) + ">\n")
        elif isinstance (node, transform.Scale):
            tv = node.evaluate ("vector", ctx.getResolvedAttributes ())
            ctx.writer.write ("scale <" + str (tv[0]) + ", " + str (tv[2]) + ", " + str (tv[1]) + ">\n")
        else:
            assert False, "unexpected transform type"


    def pushTransform(self, node, ctx):
        """
        Push a transform to the stack.
        """
        assert len (self.stacks) > 0
        self.stacks[len (self.stacks) - 1].append (node)


    def popTransform(self, node, ctx):
        """
        Pop a transform from the stack.
        """
        assert len (self.stacks) > 0
        oldNode = self.stacks[len (self.stacks) - 1].pop ()


class PovRayWriter (tree.Visitor):
    # TODO: try to reduce building of objects, seems to be slow on rendering...
    """
    Write CSG Tree to PovRay script.
    """
    def __init__(self):
        super ().__init__()
        self._rule (tree.Generator, enter = self.generator)
        self._rule (solid.Sphere, enter = self.sphere)
        self._rule (solid.Cube, enter = self.cube)
        self._rule (solid.Cylinder, enter = self.cylinder)
        self._rule (PovSolidPrism, enter = self.prism)
        self._rule (PovSolidExtrudedTorus, enter = self.torus)
        self._rule (solid.Polyhedron, enter = self.polyhedron)
        self._rule (transform.Translate, enter = self.enterTransform, leave = self.leaveTransform)
        self._rule (transform.Rotate, enter = self.enterTransform, leave = self.leaveTransform)
        self._rule (transform.Scale, enter = self.enterTransform, leave = self.leaveTransform)
        self._rule (boolean.Union, enter = self.enterUnion, leave = self.leaveBoolean)
        self._rule (boolean.Difference, enter = self.enterDifference, leave = self.leaveBoolean)
        self._rule (boolean.Intersection, enter = self.enterIntersection, leave = self.leaveBoolean)
        self._rule (tree.Group, enter = self.enterGroup, leave = self.leaveGroup)
        self._rule (part.Part, enter = self.enterPart, leave = self.leavePart)
        self._rule (scene.PointLight, enter = self.enterPointLight, leave = self.leaveLight)
        self._rule (scene.PovCameraRefinement, enter = self.enterSkipScene, leave = self.leaveSkipScene)
        self._rule (scene.PhotonSettings, enter = self.enterSceneElement, leave = self.leaveSkipScene)
        self._rule (scene.RadiositySettings, enter = self.enterSceneElement, leave = self.leaveSkipScene)


    def execute(self, filename: str, node: object, defaultAttributes: dict = None, cameraVector: list = None, width = 1, height = 1) -> None:
        """
        Run writer on tree, generate items before writing.
        """
        ctx = self.createContext (filename)
        ctx.transformStack = PovRayTransformStack ()
        if defaultAttributes == None:
            ctx.defaultArguments = tree.DefaultAttributes
        else:
            ctx.defaultArguments = defaultAttributes
        ctx.cameraVector = cameraVector
        ctx.width = width
        ctx.height = height
        includeFiles = PovRayIncludeCollector.execute (node, ctx.defaultArguments)
        self.writeIncludes (ctx, includeFiles)
        cameraRefinements = PovRayCameraRefinementCollector.execute (node, ctx.defaultArguments)
        self.writeCameraVector (ctx, cameraRefinements)
        PovRaySceneWriter.execute (node, ctx)
        node = reducegroups.ReduceGroups.execute (node, False)
        node = PovRayLinearExtrudeRewriter.execute (node, ctx.defaultArguments)
        node = PovRayRotateExtrudeRewriter.execute (node, ctx.defaultArguments)
        self.visit (node, ctx, False)
        ctx.writer.close ()


    def _enterDefault(self, node: object, context: object) -> bool:
        """
        Catch nodes without matching rule.
        """
        assert False, "node type not handled"


    def createContext(self, filename: str):
        """
        Create context for writer.
        """
        ctx = tree.VisitorContext ()
        ctx.filename = filename
        ctx.writer = textwriter.TextWriter (textwriter.FileWriter (filename))
        ctx.colorStack = []
        return ctx


    def writeIncludes(self, ctx, includeFiles: list):
        """
        Write required include files.
        """
        for i in includeFiles:
            ctx.writer.write ("#include \"" + i + "\"\n")


    def writeCameraVector(self, ctx, cameraRefinements):
        """
        Write camera vector.
        """
        if ctx.cameraVector == None:
            cv = [50, 50, 0, 0, 0, 0]
        else:
            cv = ctx.cameraVector
            assert cmath.isVector (ctx.cameraVector), "Camera vector has invalid format"

        # check which camera format is used:
        #   [location X, Y Z, lookat X, Y, Z]
        #   [translate X, Y, Z, rotate X, Y, T, distance]
        if len (ctx.cameraVector) == 6:
            locationVector = (cv[0], cv[1], cv[2])
            lookAtVector = (cv[3], cv[4], cv[5])
            rotateVector = (0, 0, 0)
            translateVector = (0, 0, 0)
        elif len (ctx.cameraVector) == 7:
            locationVector = (0, 0, cv[6])
            lookAtVector = (0, 0, 0)
            rotateVector = (-cv[3], -cv[4], -cv[5])
            translateVector = [cv[0], cv[1], cv[2]] 
        else:
            assert False, "Camera vector has invalid format"

        # write camera format
        ctx.writer.write ("camera {\n")
        ctx.writer.increaseIndent ()
        aspectRatio = (ctx.width / ctx.height)
        angle = (23.0 - 0.85 * aspectRatio) * aspectRatio
        ctx.writer.write ("angle " + str (angle) + "\n")
        ctx.writer.write ("right <" + str (aspectRatio) + ", 0, 0>\n")
        ctx.writer.write ("up <0, 1, 0>\n")
        ctx.writer.write ("location <" + str (locationVector[0]) + ", " + str (locationVector[2]) + ", " + str (locationVector[1]) + ">\n")
        ctx.writer.write ("look_at <" + str (lookAtVector[0]) + ", " + str (lookAtVector[2]) + ", " + str (lookAtVector[1]) + ">\n")
        ctx.writer.write ("rotate <" + str (rotateVector[0]) + ", 0, 0>\n") # x
        ctx.writer.write ("rotate <0, 0, " + str (rotateVector[1]) + ">\n") # y
        ctx.writer.write ("rotate <0, " + str (rotateVector[2]) + ", 0>\n") # z
        ctx.writer.write ("translate <" + str (translateVector[0]) + ", " + str (translateVector[2]) + ", " + str (translateVector[1]) + ">\n")
        ctx.writer.write (cameraRefinements)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")


    def writeTexture(self, node, ctx):
        """
        Write texture data.
        """
        # TODO: encapsulate into texture block

        # get color attribute
        color = ctx.getAttribute ("material.color")
        if color != None:
            assert isinstance (color, material.Color)

        # format pigment setting
        pigment = ctx.getAttribute ("material.pigment")
        if (pigment == None) or (pigment == ""):
            pigment = color
        if (pigment != None) and (pigment != ""):
            ctx.writer.write ("pigment {\n")
            ctx.writer.increaseIndent ()
            if isinstance (pigment, material.Color):
                # write color as pigment
                ctx.writer.write (self.formatColor (pigment) + "\n")
            else:
                # write pigment
                ctx.writer.write (str (pigment) + "\n")
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # format normal setting
        normal = ctx.getAttribute ("material.normal")
        if (normal != None) and (normal != ""):
            ctx.writer.write ("normal {\n")
            ctx.writer.increaseIndent ()
            ctx.writer.write (str (normal) + "\n")
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # finish settings
        finishedOpened = False

        # format ambient setting
        if self.writeColorOrFloat (ctx, "material.ambient", "ambient", "finish", finishedOpened, 0.2):
            finishedOpened = True

        # format emission setting
        if self.writeColor (ctx, "material.emission", "emission", "finish", finishedOpened, ""):
            finishedOpened = True

        # format diffuse setting
        if self.writeFloat (ctx, "material.diffuse", "diffuse", "finish", finishedOpened, 0.6, albedoFlag="material.diffuse.albedo"):
            finishedOpened = True

        # format brilliance setting
        if self.writeFloat (ctx, "material.brilliance", "brilliance", "finish", finishedOpened, 1):
            finishedOpened = True

        # format phong setting
        if self.writeFloat (ctx, "material.phong", "phong", "finish", finishedOpened, 0, albedoFlag="material.phong.albedo"):
            finishedOpened = True

        # format phongsize setting
        if self.writeFloat (ctx, "material.phongsize", "phong_size", "finish", finishedOpened, 40):
            finishedOpened = True

        # format specular setting
        if self.writeFloat (ctx, "material.specular", "specular", "finish", finishedOpened, 0, albedoFlag="material.specular.albedo"):
            finishedOpened = True

        # format conserve energy setting
        if self.writeBool (ctx, "material.conserveenergy", "conserve_energy", "finish", finishedOpened, False):
            finishedOpened = True

        # format useaplha energy setting
        if self.writeBool (ctx, "material.usealpha", "use_alpha", "finish", finishedOpened, False):
            finishedOpened = True

        # write reflection parameters
        refAttr = ctx.getAttribute ("material.reflection")
        reflectionWritten = False
        if (refAttr != None) and (refAttr != "") and (refAttr != 0):
            reflectionWritten = True
            if finishedOpened == False:
                ctx.writer.write ("finish {\n")
                ctx.writer.increaseIndent ()
                finishedOpened = True
            assert isinstance (refAttr, (int, float, material.Color))
            refMinAttr = ctx.getAttribute ("material.reflection.min")
            if (refMinAttr != None) and (refMinAttr != ""):
                assert isinstance (refAttr, (int, float, material.Color))
            else:
                refMinAttr = None
            ctx.writer.write ("reflection {\n")
            ctx.writer.increaseIndent ()
            if refMinAttr != None:
                ctx.writer.write (self.formatColorFloat (refMinAttr) + ", " + self.formatColorFloat (refAttr) + "\n")
            else:
                ctx.writer.write (self.formatColorFloat (refAttr) + "\n")
            self.writeFloat (ctx, "material.reflection.fresnel", "fresnel", "reflection", True, 0)
            self.writeFloat (ctx, "material.reflection.falloff", "falloff", "reflection", True, 1)
            self.writeFloat (ctx, "material.reflection.exponent", "exponent", "reflection", True, 1)
            if ctx.getAttribute ("material.reflection.metallic") in (None, ""):
                self.writeFloat (ctx, "material.metallic", "metallic", "reflection", True, 0)
            else:
                self.writeFloat (ctx, "material.reflection.metallic", "metallic", "reflection", True, 0)
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")


        # format metallic setting
        if reflectionWritten == False:
            if self.writeFloat (ctx, "material.metallic", "metallic", "finish", finishedOpened, 0):
                finishedOpened = True

        # format roughness setting
        if self.writeFloat (ctx, "material.roughness", "roughness", "finish", finishedOpened, 0.05):
            finishedOpened = True

        # format crand setting
        if self.writeFloat (ctx, "material.crand", "crand", "finish", finishedOpened, 0):
            finishedOpened = True

        # iridiscene
        iriAttr = ctx.getAttribute ("material.iridescene")
        if not (iriAttr in (None, "", 0)):
            assert isinstance (iriAttr, (int, float))
            if finishedOpened == False:
                ctx.writer.write ("finish {\n")
                ctx.writer.increaseIndent ()
                finishedOpened = True
            ctx.writer.write ("irid {\n")
            ctx.writer.increaseIndent ()
            ctx.writer.write (str (iriAttr) + "\n")
            self.writeFloat (ctx, "material.iridescene.thickness", "thickness", "irid", True, 0)
            self.writeFloat (ctx, "material.iridescene.turbulence", "turbulence", "irid", True, 0)
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # end finish settings
        if finishedOpened:
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # write interior
        interiorOpened = False

        if self.writeFloat (ctx, "material.interior.ior", "ior", "interior", interiorOpened, 1):
            interiorOpened = True

        if self.writeFloat (ctx, "material.interior.caustics", "caustics", "interior", interiorOpened, 0):
            interiorOpened = True

        if self.writeFloat (ctx, "material.interior.dispersion", "dispersion", "interior", interiorOpened, 1):
            interiorOpened = True

        if self.writeInteger (ctx, "material.interior.dispersionsamples", "dispersion_samples", "interior", interiorOpened, 7):
            interiorOpened = True

        if self.writeFloat (ctx, "material.interior.fadedistance", "fade_distance", "interior", interiorOpened, 0):
            interiorOpened = True

        if self.writeFloat (ctx, "material.interior.fadepower", "fade_power", "interior", interiorOpened, 0):
            interiorOpened = True

        if self.writeColor (ctx, "material.interior.fadecolor", "fade_color", "interior", interiorOpened, ""):
            interiorOpened = True

        if interiorOpened:
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # write photon mapping settings
        photonsOpened = False

        photonsTarget = ctx.getAttribute ("material.photons.target")
        if photonsTarget == True:
            if photonsOpened == False:
                ctx.writer.write ("photons {\n")
                ctx.writer.increaseIndent ()
                photonsOpened = True
            targetValStr = ""
            photonsTargetValue = ctx.getAttribute ("material.photons.targetvalue")
            if not photonsTargetValue in (None, 1.0):
                targetValStr = " " + str (photonsTargetValue)
            ctx.writer.write ("target" + targetValStr + "\n")


        if self.writeBool (ctx, "material.photons.refraction", "refraction", "photons", photonsOpened, False):
            photonsOpened = True

        if self.writeBool (ctx, "material.photons.reflection", "reflection", "photons", photonsOpened, False):
            photonsOpened = True

        if self.writeFloat (ctx, "material.photons.density", "density", "photons", photonsOpened, 1):
            photonsOpened = True

        if self.writeBool (ctx, "material.photons.collect", "collect", "photons", photonsOpened, True):
            photonsOpened = True

        if self.writeBool (ctx, "material.photons.passthrough", "passthrough", "photons", photonsOpened, False):
            photonsOpened = True

        if self.writeBool (ctx, "material.photons.arealight", "arealight", "photons", photonsOpened, False):
            photonsOpened = True

        if self.writeBool (ctx, "material.photons.splitunion", "splitunion", "photons", photonsOpened, True):
            photonsOpened = True

        if photonsOpened:
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")


    def formatColorFloat(self, value):
        if isinstance (value, material.Color):
            return self.formatColor (value)
        if isinstance (value, (int, float)):
            return str (value)
        return ""


    def writeColor(self, ctx, attributeName, povName, group: str = None, groupOpened: bool = False, default = None):
        """
        Write color as attribute.
        """
        attr = ctx.getAttribute (attributeName)
        if (attr != None) and (attr != ""):
            if isinstance (attr, material.Color):
                # use color as attribute
                if (default == None) or (attr != default):
                    if groupOpened == False:
                        ctx.writer.write (group + " {\n")
                        ctx.writer.increaseIndent ()
                        groupOpened = True
                    ctx.writer.write (povName + " " + self.formatColor (attr) + "\n")
            else:
                assert False, "unexpected argument type"
        return groupOpened


    def writeColorOrFloat(self, ctx, attributeName, povName, group: str = None, groupOpened: bool = False, default = None):
        """
        Write color or float value as attribute.
        """
        attr = ctx.getAttribute (attributeName)
        if (attr != None) and (attr != ""):
            if isinstance (attr, material.Color):
                # use color as attribute
                if (default == None) or (attr != default):
                    if groupOpened == False:
                        ctx.writer.write (group + " {\n")
                        ctx.writer.increaseIndent ()
                        groupOpened = True
                    ctx.writer.write (povName + " " + self.formatColor (attr) + "\n")
            elif isinstance (attr, (int, float)):
                if (default == None) or (attr != default):
                    # use factor as attribute
                    if groupOpened == False:
                        ctx.writer.write (group + " {\n")
                        ctx.writer.increaseIndent ()
                        groupOpened = True
                    ctx.writer.write (povName + " " + str (attr) + "\n")
                else:
                    # ambient default value
                    pass
            else:
                assert False, "unexpected argument type"
        return groupOpened


    def writeInteger(self, ctx, attributeName, povName, group: str = None, groupOpened: bool = False, default = None):
        """
        Write integer value as attribute.
        """
        attr = ctx.getAttribute (attributeName)
        if (attr != None) and (attr != ""):
            if isinstance (attr, int):
                if (default == None) or (attr != default):
                    if groupOpened == False:
                        ctx.writer.write (group + " {\n")
                        ctx.writer.increaseIndent ()
                        groupOpened = True
                    ctx.writer.write (povName + " " + str (attr) + "\n")
            else:
                assert False, "unexpected argument type"
        return groupOpened


    def writeFloat(self, ctx, attributeName, povName, group: str = None, groupOpened: bool = False, default = None, albedoFlag = None):
        """
        Write float value as attribute.
        """
        albedoStr = ""
        albedoAttr = ctx.getAttribute (albedoFlag)
        if albedoAttr == True:
            albedoStr = " albedo"
        attr = ctx.getAttribute (attributeName)
        if (attr != None) and (attr != ""):
            if isinstance (attr, (int, float)):
                if (default == None) or (attr != default):
                    if groupOpened == False:
                        ctx.writer.write (group + " {\n")
                        ctx.writer.increaseIndent ()
                        groupOpened = True
                    ctx.writer.write (povName + albedoStr + " " + str (attr) + "\n")
            else:
                assert False, "unexpected argument type"
        return groupOpened


    def writeBool(self, ctx, attributeName, povName, group: str = None, groupOpened: bool = False, default = None):
        """
        Write boolean value as attribute.
        """
        attr = ctx.getAttribute (attributeName)
        if (attr != None) and (attr != ""):
            if isinstance (attr, bool):
                if (default == None) or (attr != default):
                    if groupOpened == False:
                        ctx.writer.write (group + " {\n")
                        ctx.writer.increaseIndent ()
                        groupOpened = True
                    val = "on" if attr else "off"
                    ctx.writer.write (povName + " " + val + "\n")
            else:
                assert False, "unexpected argument type"
        return groupOpened


    def formatColor(self, color):
        """
        Format color item.
        """
        assert isinstance (color, material.Color)
        return color.asString ()


    def enterSceneElement(self, node, ctx):
        """
        Write scene element settings
        """
        ctx.writer.write (node.format ())


    # handle generator node

    def generator(self, node, context):
        self.visit (node.generate (context.getResolvedAttributes ()), context)
        return True


    # handle solids

    def sphere(self, node, ctx):
        """
        Enter a sphere.
        """
        ctx.writer.write ("sphere {\n")
        ctx.writer.increaseIndent ()
        radius = node.evaluate ("radius", ctx.getResolvedAttributes ())
        ctx.writer.write ("<0, 0, 0>, " + str (radius) + "\n")
        self.writeTexture (node, ctx)
        ctx.transformStack.pushLayer ()
        ctx.transformStack.popAndWriteLayer (ctx)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")
        return False


    def cube(self, node, ctx):
        """
        Enter a cube.
        """
        ctx.writer.write ("box {\n")
        ctx.writer.increaseIndent ()
        size = node.evaluate ("size", ctx.getResolvedAttributes ())
        sz = [-size[0] / 2, -size[1] / 2, -size[2] / 2, size[0] / 2, size[1] / 2, size[2] / 2]
        ctx.writer.write ("<" + str (sz[0]) + ", " + str (sz[2]) + ", " + str (sz[1]) + ">" + "\n")
        ctx.writer.write ("<" + str (sz[3]) + ", " + str (sz[5]) + ", " + str (sz[4]) + ">" + "\n")
        self.writeTexture (node, ctx)
        ctx.transformStack.pushLayer ()
        ctx.transformStack.popAndWriteLayer (ctx)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")
        return False


    def cylinder(self, node, ctx):
        """
        Enter a cylinder.
        """
        ctx.writer.write ("cone {\n")
        ctx.writer.increaseIndent ()
        height = node.evaluate ("height", ctx.getResolvedAttributes ())
        radius = node.evaluate ("radius", ctx.getResolvedAttributes ())
        radius2 = node.evaluate ("radius2", ctx.getResolvedAttributes ())
        if radius2 == None:
            radius2 = radius
        ctx.writer.write ("<0, " + str (-height / 2.0) + ", 0>, " + str (radius) + "\n")
        ctx.writer.write ("<0, " + str (height / 2.0) + ", 0>, " + str (radius2) + "\n")
        self.writeTexture (node, ctx)
        ctx.transformStack.pushLayer ()
        ctx.transformStack.popAndWriteLayer (ctx)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")
        return False


    def prism(self, node, ctx):
        """
        Enter a prism.
        """
        ctx.writer.write ("prism {\n")
        ctx.writer.increaseIndent ()
        height = node.height
        polygon = node.polygon
        type_ = node.type_
        points = polygon.getData ()

        # here we go, got height and points, write the prism
        ctx.writer.write (type_ + "\n")
        h1 = (height * -0.5) + node.centerOffset
        h2 = (height * 0.5) + node.centerOffset
        pointCount = len (points)
        if type_ == "linear_spline":
            outPointCount = pointCount
        elif type_ == "quadratic_spline":
            outPointCount = len (points) + 1
        elif type_ == "cubic_spline":
            outPointCount = len (points) + 2
        elif type_ == "bezier_spline":
            outPointCount = pointCount
        else:
            assert False, "spline type not implemented"
        ctx.writer.write (str (h1) + ", " + str (h2) + ", " + str (outPointCount) + ",\n")

        # write points
        if type_ in ("linear_spline", "quadratic_spline", "cubic_spline"):
            if type_ in ("quadratic_spline", "cubic_spline"):
                ctx.writer.write ("<" + str (points[-1][0]) + ", " + str (points[-1][1]) + ">")
            if type_ in ("linear_spline", "quadratic_spline"):
                for pointId in range (0, pointCount):
                    point = points[pointId]
                    ctx.writer.write ("<" + str (point[0]) + ", " + str (point[1]) + ">")
                    if pointId < (pointCount - 1):
                        ctx.writer.write(",\n")
                    else:
                        ctx.writer.write("\n")
            else:
                for pointId in range (0, pointCount):
                    point = points[pointId]
                    ctx.writer.write ("<" + str (point[0]) + ", " + str (point[1]) + ">,\n")
                ctx.writer.write ("<" + str (points[0][0]) + ", " + str (points[0][1]) + ">\n")
        elif type_ == "bezier_spline":
            for pointId in range (0, pointCount):
                point = points[pointId]
                ctx.writer.write ("<" + str (point[0]) + ", " + str (point[1]) + ">")
                if pointId < (pointCount - 1):
                    ctx.writer.write(",\n")
                else:
                    ctx.writer.write("\n")
        else:
            assert False, "invalid spline type"

        # end wiring prism
        self.writeTexture (node, ctx)
        ctx.transformStack.pushLayer ()
        ctx.transformStack.popAndWriteLayer (ctx)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")
        return False


    def torus(self, node, ctx):
        """
        Enter a torus.
        """
        ctx.writer.write ("lathe {\n")
        ctx.writer.increaseIndent ()
        numPoints = len (node.points)
        if node.type_ == "bezier_spline":
            ctx.writer.write ("bezier_spline\n")
            ctx.writer.write (str (numPoints))
        elif node.type_ == "linear_spline":
            ctx.writer.write ("linear_spline\n")
            ctx.writer.write (str (numPoints + 1))
        elif node.type_ == "quadratic_spline":
            ctx.writer.write ("quadratic_spline\n")
            ctx.writer.write (str (numPoints + 2))
        elif node.type_ == "cubic_spline":
            ctx.writer.write ("cubic_spline\n")
            ctx.writer.write (str (numPoints + 3))
        else:
            assert False, "TODO: not handled yet"

        # write first control point
        if node.type_ in ("quadratic_spline", "cubic_spline"):
            ctx.writer.write (",\n<" + str (node.points[-1][0]) + ", " + str (node.points[-1][1]) + ">")

        # write points
        for point in node.points:
            ctx.writer.write (",\n<" + str (point[0]) + ", " + str (point[1]) + ">")

        # close polygon points
        if node.type_ in ("linear_spline", "quadratic_spline"):
            ctx.writer.write (",\n<" + str (node.points[0][0]) + ", " + str (node.points[0][1]) + ">\n")
        elif node.type_ == "cubic_spline":
            ctx.writer.write (",\n<" + str (node.points[0][0]) + ", " + str (node.points[0][1]) + ">")
            ctx.writer.write (",\n<" + str (node.points[1][0]) + ", " + str (node.points[1][1]) + ">\n")
        else:
            ctx.writer.write ("\n")

        # end wiring torus
        self.writeTexture (node, ctx)
        ctx.transformStack.pushLayer ()
        ctx.transformStack.popAndWriteLayer (ctx)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")
        return False


    def polyhedron(self, node, ctx):
        """
        Enter a polyhedron.
        """
        points, faces, normals, uvVectors = node.getData ()
        ctx.writer.write ("mesh2 {\n")
        ctx.writer.increaseIndent ()

        facesWithNormals = []
        facesWithoutNormals = []

        for face in faces:
            if len (face) >= 6:
                if (face[3] != None) or (face[4] != None) or (face[5] != None):
                    facesWithNormals.append (face)
                else:
                    facesWithoutNormals.append (face)
            else:
                facesWithoutNormals.append (face)
        processNormals = False
        if len (facesWithNormals) > 0:
            if normals != None:
                if len (normals) > 0:
                    processNormals = True

        # check if we need to process uv vectors
        processUvVectors = False
        if uvVectors != None:
            if len (uvVectors) > 0:
                for face in faces:
                    if len (face) >= 9:
                        if (face[6] != None) or (face[7] != None) or (face[8] != None):
                            processUvVectors = True

        # write points to mesh2 structure
        ctx.writer.write ("vertex_vectors {\n")
        ctx.writer.increaseIndent ()
        ctx.writer.write (str (len (points)) + ",\n")
        for pid in range (0, len (points)):
            point = points[pid]
            s = self.formatVector (point)
            if pid < (len (points) - 1):
                s = s + ","
            ctx.writer.write (s + "\n")
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")

        # write list of normals
        if processNormals:
            ctx.writer.write ("normal_vectors {\n")
            ctx.writer.increaseIndent ()
            ctx.writer.write (str (len (normals)) + ",\n")
            for nid in range (0, len (normals)):
                normal = normals[nid]
                s = self.formatVector (normal)
                if nid < (len (normals) - 1):
                    s = s + ","
                ctx.writer.write (s + "\n")
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # write list of uv vectors
        if processUvVectors:
            ctx.writer.write ("uv_vectors {\n")
            ctx.writer.increaseIndent ()
            ctx.writer.write (str (len (uvVectors)) + ",\n")
            for uid in range (0, len (uvVectors)):
                uv = uvVectors[uid]
                s = self.formatUvVector (uv)
                if uid < (len (uvVectors) - 1):
                    s = s + ","
                ctx.writer.write (s + "\n")
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # write faces
        ctx.writer.write ("face_indices {\n")
        ctx.writer.increaseIndent ()
        allFaces = facesWithNormals + facesWithoutNormals
        ctx.writer.write (str (len (allFaces)) + ",\n")
        for fid in range (0, len (allFaces)):
            face = allFaces[fid]
            s = self.formatVectorIndices (face[0], face[1], face[2])
            if fid < (len (allFaces) - 1):
                s = s + ","
            ctx.writer.write (s + "\n")
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")

        # write face normals
        if processNormals:
            ctx.writer.write ("normal_indices {\n")
            ctx.writer.increaseIndent ()
            ctx.writer.write (str (len (facesWithNormals)) + ",\n")
            for fid in range (0, len (facesWithNormals)):
                face = facesWithNormals[fid]
                s = self.formatVectorIndices (face[3], face[4], face[5])
                if fid < (len (facesWithNormals) - 1):
                    s = s + ","
                ctx.writer.write (s + "\n")
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # write face uv mappings
        if processUvVectors:
            ctx.writer.write ("uv_indices {\n")
            ctx.writer.increaseIndent ()
            ctx.writer.write (str (len (allFaces)) + ",\n")
            for fid in range (0, len (allFaces)):
                face = allFaces[fid]
                if len (face) < 9:
                    s = "<0, 0, 0>"
                else:
                    s = self.formatVectorIndices (face[6], face[7], face[8])
                if fid < (len (allFaces) - 1):
                    s = s + ","
                ctx.writer.write (s + "\n")
            ctx.writer.decreaseIndent ()
            ctx.writer.write ("}\n")

        # write inside vector
        ctx.writer.write ("inside_vector <0, 0, 1>\n")    

        # end of mash
        self.writeTexture (node, ctx)
        ctx.transformStack.pushLayer ()
        ctx.transformStack.popAndWriteLayer (ctx)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")


    def formatVector(self, vector):
        """
        Format a 3 dimensional vector.
        """
        return "<" + str (vector[0]) + ", " + str (vector[2]) + ", " + str (vector[1]) + ">"


    def formatVectorIndices(self, a, b, c):
        """
        Format vector indices.
        """
        astr = str (a) if a != None else 0
        bstr = str (b) if b != None else 0
        cstr = str (c) if c != None else 0
        return "<" + astr + ", " + bstr + ", " + cstr + ">"


    # handle transforms

    def enterGroup(self, node, ctx):
        """
        Enter an object.
        """
        return True


    def leaveGroup(self, node, ctx):
        """
        Leave a group.
        """
        return node


    def enterTransform(self, node, ctx):
        """
        Enter a transform operation.
        """
        ctx.transformStack.pushTransform (node, ctx)
        return True


    def leaveTransform(self, node, ctx):
        """
        Leave a transform operation.
        """
        ctx.transformStack.popTransform (node, ctx)
        return node


    # handle boolean operations

    def enterUnion(self, node, ctx):
        """
        Enter an union.
        """
        ctx.writer.write ("union {\n")
        ctx.writer.increaseIndent ()
        ctx.transformStack.pushLayer ()
        return True


    def enterDifference(self, node, ctx):
        """
        Enter an difference.
        """
        ctx.writer.write ("difference {\n")
        ctx.writer.increaseIndent ()
        ctx.transformStack.pushLayer ()
        return True


    def enterIntersection(self, node, ctx):
        """
        Enter an intersection.
        """
        ctx.writer.write ("intersection {\n")
        ctx.writer.increaseIndent ()
        ctx.transformStack.pushLayer ()
        return True


    def leaveBoolean(self, node, ctx):
        """
        Leave a boolean operation.
        """
        ctx.transformStack.popAndWriteLayer (ctx)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")
        return node


    # handle parts

    def enterPart(self, node, ctx):
        return True
        

    def leavePart(self, node, ctx):
        return node
            

    def enterSkipScene(self, node, ctx):
        return True
        

    def leaveSkipScene(self, node, ctx):
        return node
        

    # handle ligths
    # TODO: remove extra class - handle here
    def enterPointLight(self, node, ctx):
        ctx.writer.write ("light_source {\n")
        ctx.writer.increaseIndent ()
        position = node.evaluate ("position", ctx.getResolvedAttributes ())
        color = "rgb<1,1,1>" # TODO: from attributes
        ctx.writer.write ("<" + str (position[0]) + ", " + str (position[2]) + ", " + str (position[1]) + ">, " + color)
        ctx.writer.decreaseIndent ()
        ctx.writer.write ("}\n")
        return True


    def leaveLight(self, node, ctx):
        return node



class _PovRay:
    """
    PovRay plugin.
    """
    def __init__(self):
        self.silent = True
        self.writer = PovRayWriter ()
        self.instanceLock = threading.Lock ()


    def getScriptExtension(self):
        return "pov"


    def getImageExtension(self):
        return "png"


    def writeTree(self, filename: str, tree: object, defaultArgs: dict = None, cameraVector: list = None, width = 1, height = 1) -> None:
        """
        Write CSG Tree to povray script file.
        """
        self.writer.execute (filename, tree, defaultArgs, cameraVector, width, height)
        return True


    def render(self, scriptFile, outputFile, width, height, antialias, quality, tree: object, cameraVector: list = None, preview: bool = False):
        """
        Render image with antialiasing.
        """
        args = [PovRayExecutable, "+O" + outputFile, "+W" + str (int(width)), "+H" + str (int(height)), "+FN"]
        if preview != True:
            args.append ("-D")

        # render quality settings
        args.append ("+Q" + str (quality))

        # antialiasing settings
        anitaliasThresholds = [
            "+A0.2",
            "+A0.1",
            "+A0.05",
            "+A0.025",
            "+A0.0125",
            "+A0.00625",
            "+A0.05",
            "+A0.025",
            "+A0.0125",
            "+A0.00625",
            "+A0.003",
            "+A0.00015"
        ]
        if antialias > 0:
            if antialias >= 6:
                # adaptive
                args.append ("+AM2")
            else:
                # non adaptive
                args.append ("+AM1")
            args.append (anitaliasThresholds[antialias])

        args.append (scriptFile)
        # let povray instance not run in parallel!
        self.instanceLock.acquire ()
        cmd = subprocess.run (args, capture_output=self.silent)
        if self.silent and (cmd.returncode != 0):
            print (cmd.stderr)
            print (cmd.stdout)
        self.instanceLock.release ()
        return cmd.returncode == 0


    def canManufacture(self):
        """
        Check if manufacturing is supported.
        """
        return False


    def manufacture(self, scriptFile, outputFile):
        """
        Manufacture a part.
        """
        # not supported
        return False
        
        
    def scriptRequiresRenderingDetails(self):
        """
        PovRay scripts are dependent from the camera position and output image size.
        """
        return True


    def getAttributesOverrides (self):
        """
        Returns the override attributes of this materializer.
        """
        return {
            'materializer.supportsBezierCurves': True
        }
        




# single instance of povray extension
PovRay = _PovRay ()



from . import tree
from .attributes import Attributes




class Boolean:
    """
    Informational base class of :ref:`Boolean<Boolean>` objects. This empty class is implemented for type matching purposes.
    """
    def __init__ (self):
        #: The number of dimensions is determined from the child instance.
        self.dimensions = None




class Union(tree.Node, Boolean):
    """
    Combine two or more :ref:`Shapes<Shape>` or :ref:`Solids<Solid>`.
    """
    def __init__ (self,
                    children: object = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert children are valid
        if isinstance (children, (list, tuple)):
            assert len (children) >= 2, "Expected at least two children"
            for childId in range (0, len (children)):
                assert isinstance (children[childId], tree.Item), "Expected a csg tree item as child"
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child"
            children = (children,)
    
        tree.Node.__init__ (self, children = children, name = name, attributes = attributes)
      
        # calculate number of required dimensions
        dimensions = None
        for childId in range (0, len (children)):
            child = children[childId]
            if child != None:
                if not isinstance (child, tree.Empty):
                    if dimensions == None:
                        dimensions = child.dimensions
                    else:
                        if child.dimensions != None:
                            assert child.dimensions == dimensions, "Expected child nodes to have the same number of dimensions."

        # The number of dimensions is determined from the child instance.      
        self.dimensions = dimensions
      
        #: Tuple containing at least two :ref:`Shapes<Shape>` or :ref:`Solids<Solid>`.
        self.children = self.children


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
        # TODO:
        )




class Difference(tree.Node, Boolean):
    """
    Difference of two :ref:`Shapes<Shape>` or :ref:`Solids<Solid>`.
    """
    def __init__ (self,
                    children: object = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert children are valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 2, "Expected two children"
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child"
            assert isinstance (children[1], tree.Item), "Expected a csg tree item as child"
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child"
            children = (children,)
    
        super ().__init__ (children = children, name = name, attributes = attributes)      
      
        # calculate number of dimensions
        dimensions = children[0].dimensions if children[0] != None else None
        if dimensions == None:
            dimensions = children[1].dimensions if children[1] != None else None
        else:
            if children[1] != None:
                if children[1].dimensions != None:
                    assert children[1].dimensions == dimensions, "Expected both children to have the same number of dimensions." 

        # The number of dimensions is determined from the child instance.
        self.dimensions = dimensions
      
        #: Tuple containing two :ref:`Shapes<Shape>` or :ref:`Solids<Solid>`.
        self.children = self.children


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
        )




class Intersection(tree.Node, Boolean):
    """
    Intersection of two :ref:`Shapes<Shape>` or :ref:`Solids<Solid>`.
    """
    def __init__ (self,
                    children: object = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert children are valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 2, "Expected two children"
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child"
            assert isinstance (children[1], tree.Item), "Expected a csg tree item as child"
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child"
            children = (children,)
    
        super ().__init__ (children = children, name = name, attributes = attributes)
      
        # calculate number of dimensions
        dimensions = children[0].dimensions if children[0] != None else None
        if dimensions == None:
            dimensions = children[1].dimensions if children[1] != None else None
        else:
            if children[1] != None:
                if children[1].dimensions != None:
                    assert children[1].dimensions == dimensions, "Expected both children to have the same number of dimensions." 

        # The number of dimensions is determined from the child instance.
        self.dimensions = dimensions

        #: Tuple containing two :ref:`Shapes<Shape>` or :ref:`Solids<Solid>`.
        self.children = self.children


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
        )



import subprocess
from PIL import Image

# increase maximum image buffer size
Image.MAX_IMAGE_PIXELS = Image.MAX_IMAGE_PIXELS * 4




def downsampleImage (input, output, factor):
    """
    Downsample an image
    """
    img = Image.open(input)
    w = int (float(img.size[0]) / float (factor))
    h = int (float(img.size[1]) / float (factor))
    img = img.resize((w, h), Image.ANTIALIAS)
    img.save(output) 
    return True
